import copy
import logging

import atlas_mpl_style as ampl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import auc, roc_curve

import hepynet.common.hepy_type as ht
from hepynet.evaluate import evaluate_utils
from hepynet.train import hep_model

logger = logging.getLogger("hepynet")


def calculate_auc(y_tag, y_pred, weights):
    """Returns auc of given sig/bkg array."""
    auc_value = []
    num_nodes = y_tag.shape[1]
    for node_id in range(num_nodes):
        fpr_dm, tpr_dm, _ = roc_curve(
            y_tag[:, node_id], y_pred[:, node_id], sample_weight=weights
        )
        sort_index = fpr_dm.argsort()
        fpr_dm = fpr_dm[sort_index]
        tpr_dm = tpr_dm[sort_index]
        auc_value.append(auc(fpr_dm, tpr_dm))
    return auc_value


def plot_feature_importance(
    model_wrapper: hep_model.Model_Base,
    df: pd.DataFrame,
    job_config,
    save_dir,
    max_feature=50,
):
    """Calculates importance of features and sort the feature.

    Definition of feature importance used here can be found in:
    https://christophm.github.io/interpretable-ml-book/feature-importance.html#feature-importance-data

    """
    logger.info("Plotting feature importance.")
    ic = job_config.input.clone()
    tc = job_config.train.clone()
    ac = job_config.apply.clone()
    # prepare
    model = model_wrapper.get_model()
    cols = ic.selected_features
    test_index = df["is_train"] == False
    y_test = df.loc[test_index, ["y"]].to_numpy("float32")
    y_pred = df.loc[test_index, ["y_pred"]].to_numpy("float32")
    wt_test = df.loc[test_index, "weight"].to_numpy("float32")
    all_nodes = ["sig"] + tc.output_bkg_node_names
    # Make plots
    fig_save_pattern = f"{save_dir}/importance_{{}}.png"
    num_feature = len(cols)
    if num_feature < max_feature:
        canvas_height = num_feature / 2 + 5
    else:
        canvas_height = max_feature / 2 + 5
    base_auc = calculate_auc(y_test, y_pred, wt_test)
    # Calculate importance
    feature_auc = []
    logger.info("Making predictions with shuffled one column")
    for num, feature_name in enumerate(cols):
        x_shuffle = df.loc[test_index, cols].values
        np.random.shuffle(x_shuffle[:, num])
        y_pred, _, _ = evaluate_utils.k_folds_predict(
            model, x_shuffle, silence=True
        )
        current_auc = calculate_auc(y_test, y_pred, wt_test)
        feature_auc.append(current_auc)
    for node_id, node in enumerate(all_nodes):
        logger.info(f"> making importance plot for node: {node}")
        fig_save_path = fig_save_pattern.format(node)
        fig, ax = plt.subplots(figsize=(11.111, canvas_height))
        logger.info(f"> base auc: {base_auc[node_id]}")
        feature_importance = np.zeros(num_feature)
        for num, feature_name in enumerate(cols):
            current_auc = feature_auc[num][node_id]
            feature_importance[num] = (1 - current_auc) / (
                1 - base_auc[node_id]
            )
            logger.info(f"> {feature_name} : {feature_importance[num]}")

        # Sort
        sort_list = np.argsort(feature_importance).astype(int)
        sorted_importance = feature_importance[sort_list]
        sorted_names = [cols[i] for i in sort_list]
        rank_str = " > ".join(sorted_names[::-1])
        logger.info(f"> feature importance rank: {rank_str}")
        # Plot
        if num_feature > max_feature:
            num_show = max_feature
        else:
            num_show = num_feature
        ax.barh(
            np.arange(num_show),
            sorted_importance[:num_show],
            align="center",
            alpha=0.5,
            log=ac.cfg_importance_study.log,
        )
        ax.axvline(x=1, ls="--", color="r")
        _, x_max = ax.get_xlim()
        ax.set_xlim(0.1, x_max)
        _, y_max = ax.get_ylim()
        ax.set_ylim(-0.5, y_max + 2)
        if ac.plot_atlas_label:
            ampl.plot.draw_atlas_label(
                0.05, 0.95, ax=ax, **(ac.atlas_label.get_config_dict())
            )
        ax.set_title("feature importance")
        ax.set_yticks(np.arange(num_show))
        ax.set_yticklabels(sorted_names[:num_show])
        fig.subplots_adjust(left=0.3)
        fig.savefig(fig_save_path)



# -*- coding: utf-8 -*-
"""Model wrapper class for DNN training"""
import copy
import datetime
import logging
import pathlib
from typing import Dict, Iterable, Optional, Tuple

import keras
import numpy as np
import ray
import tensorflow as tf
import yaml
from keras import backend as K
from keras.callbacks import ModelCheckpoint, callbacks
from keras.layers import Dense, Dropout
from keras.models import Sequential
from keras.optimizers import SGD, Adagrad, Adam, RMSprop
from ray import tune
from ray.tune.integration.keras import TuneReportCallback
from sklearn.metrics import roc_auc_score

import hepynet.common.hepy_type as ht
from hepynet.data_io import feed_box
from hepynet.train import train_utils

logger = logging.getLogger("hepynet")

# fix tensorflow 2.2 issue
gpus = tf.config.experimental.list_physical_devices("GPU")
if gpus:
    try:
        # Currently, memory growth needs to be the same across GPUs
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.experimental.list_logical_devices("GPU")
        logger.info(
            f"GPU availability: {len(gpus)} Physical GPUs, {len(logical_gpus)} Logical GPUs"
        )
    except RuntimeError as e:
        # Memory growth must be set before GPUs have been initialized
        logger.error(e)
# Set up mixed float16 training for TF >= 2.4 TODO: need to upgrade TF
# mixed_precision.set_global_policy('mixed_float16')

# self-defined metrics functions
def plain_acc(y_true, y_pred):
    return K.mean(K.less(K.abs(y_pred * 1.0 - y_true * 1.0), 0.5))
    # return 1-K.mean(K.abs(y_pred-y_true))


class Model_Base(object):
    """Base model of deep neural network
    """

    def __init__(self, job_config: ht.config):
        self._job_config = job_config.clone()
        self._model_create_time = str(datetime.datetime.now())
        self._model_is_compiled = False
        self._model_is_loaded = False
        self._model_is_saved = False
        self._model_is_trained = False
        self._model_name = self._job_config.train.model_name
        self._model_save_path = None
        self._train_history = list()

        num_folds = self._job_config.train.k_folds
        if isinstance(num_folds, int) and num_folds >= 2:
            self._num_folds = num_folds
        else:
            self._num_folds = 1

    def build(self):
        logger.warn(
            "The virtual function Model_Base.build() is called, please implement override funcion!"
        )

    def get_feedbox(self) -> feed_box.Feedbox:
        return self._feedbox

    def get_job_config(self) -> ht.config:
        return self._job_config

    def get_model(self, fold_num=None):
        """Returns model."""
        if not self._model_is_compiled:
            logger.warning("Model is not compiled")
        if fold_num is None:
            return self._model
        else:
            return self._model[fold_num]

    def get_model_meta(self) -> dict:
        return self._model_meta

    def get_model_save_dir(
        self, fold_num: Optional[int] = None
    ) -> ht.pathlike:
        save_dir = f"{self._job_config.run.save_sub_dir}/models"
        if fold_num is not None:
            save_dir += f"/fold_{fold_num}"
        return save_dir

    def set_inputs(self, job_config: ht.config):
        """Prepare feedbox to generate inputs"""
        rc = self._job_config.run.clone()
        try:
            input_dir = pathlib.Path(rc.save_sub_dir) / "models"
            with open(input_dir / "norm_dict.yaml", "r") as norm_file:
                norm_dict = yaml.load(norm_file, Loader=yaml.UnsafeLoader)
            logger.info(f"Successfully loaded norm_dict in {input_dir}")
        except:
            norm_dict = None
        feedbox = feed_box.Feedbox(job_config, norm_dict=norm_dict)
        self._model_meta["norm_dict"] = copy.deepcopy(feedbox.get_norm_dict())
        self._feedbox = feedbox
        self._array_prepared = feedbox._array_prepared

    def load_model(self, epoch: Optional[int] = None):
        """Loads saved model."""
        # load model(s)
        self._model = list()
        num_exist_models = 0
        for fold_num in range(self._num_folds):
            model_dir = self.get_model_save_dir(fold_num=fold_num)
            if epoch is None:
                model_path = pathlib.Path(f"{model_dir}/{self._model_name}.h5")
            else:
                model_path = pathlib.Path(
                    f"{model_dir}/{self._model_name}_epoch{epoch}.h5"
                )
            if model_path.exists():
                fold_model = keras.models.load_model(
                    model_path,
                    custom_objects={"plain_acc": plain_acc,},
                    compile=False,
                )  # it's important to specify custom_objects
                self._model.append(fold_model)
                num_exist_models += 1
        self._model_is_loaded = True
        # Load parameters
        model_dir = self.get_model_save_dir()
        self.load_model_parameters(model_dir)
        self.model_paras_is_loaded = True
        if epoch is None:
            logger.info(f"{num_exist_models}/{self._num_folds} models loaded")
        else:
            logger.info(
                f"{num_exist_models}/{self._num_folds} models loaded for epoch {epoch}"
            )

    def load_model_parameters(self, model_dir: ht.pathlike):
        """Retrieves model parameters from yaml file."""
        paras_path = f"{model_dir}/fold_{0}/{self._model_name}_paras.yaml"
        with open(paras_path, "r") as paras_file:
            paras_dict = yaml.load(paras_file, Loader=yaml.UnsafeLoader)
        # update meta data
        model_meta_save = paras_dict["model_meta"]
        self._model_meta = model_meta_save
        self._model_name = model_meta_save["model_name"]
        self._model_label = model_meta_save["model_label"]
        self._model_note = model_meta_save["model_note"]
        self._model_create_time = model_meta_save["model_create_time"]
        self._model_is_compiled = model_meta_save["model_is_compiled"]
        self._model_is_saved = model_meta_save["model_is_saved"]
        self._model_is_trained = model_meta_save["model_is_trained"]
        # load train history
        for fold_num in range(self._num_folds):
            paras_path = (
                f"{model_dir}/fold_{fold_num}/{self._model_name}_paras.yaml"
            )
            with open(paras_path, "r") as paras_file:
                fold_paras_dict = yaml.load(
                    paras_file, Loader=yaml.UnsafeLoader
                )
            self._train_history[fold_num] = fold_paras_dict["train_history"]

    def save_model(
        self, file_name: Optional[str] = None, fold_num: Optional[int] = None
    ):
        """Saves trained model."""
        # Define save path
        save_dir = self.get_model_save_dir(fold_num=fold_num)
        if file_name is None:
            file_name = self._model_name
        # Check path
        save_path = f"{save_dir}/{file_name}.h5"
        pathlib.Path(save_dir).mkdir(parents=True, exist_ok=True)
        # Save
        if fold_num is not None:
            self._model[fold_num].save(save_path)
        else:
            self._model[0].save(save_path)
        self._model_save_path = save_path
        logger.debug(
            f"model: {self._model_name} has been saved to: {save_path}"
        )
        self._model_is_saved = True

    def save_model_paras(
        self, file_name: Optional[str] = None, fold_num: Optional[int] = None
    ):
        """Save model parameters to yaml file."""
        rc = self._job_config.run.clone()
        # prepare paras
        paras_dict = dict()
        model_meta_save = copy.deepcopy(self._model_meta)
        model_meta_save["model_name"] = self._model_name
        model_meta_save["model_label"] = self._model_label
        model_meta_save["model_note"] = self._model_note
        model_meta_save["model_create_time"] = self._model_create_time
        model_meta_save["model_is_compiled"] = self._model_is_compiled
        model_meta_save["model_is_saved"] = self._model_is_saved
        model_meta_save["model_is_trained"] = self._model_is_trained
        paras_dict["model_meta"] = model_meta_save
        paras_dict["train_history"] = self._train_history[fold_num]
        # save to file
        save_dir = self.get_model_save_dir(fold_num=fold_num)
        if file_name is None:
            file_name = self._model_name
        save_path = f"{save_dir}/{file_name}_paras.yaml"
        with open(save_path, "w") as write_file:
            yaml.dump(paras_dict, write_file, indent=2)
        logger.debug(f"model parameters has been saved to: {save_path}")

        norm_dict_path = (
            pathlib.Path(rc.save_sub_dir) / "models" / "norm_dict.yaml"
        )
        with open(norm_dict_path, "w") as norm_file:
            yaml.dump(self._feedbox.get_norm_dict(), norm_file, indent=2)


class Model_Sequential_Base(Model_Base):
    """Sequential model base.

    Note:
        This class should not be used directly
    """

    def __init__(self, job_config: ht.config):
        super().__init__(job_config)
        tc = self._job_config.train
        ic = self._job_config.input
        # Model parameters
        self._model_label = "mod_seq_base"
        self._model_note = "Basic sequential model."
        self._model_input_dim = len(ic.selected_features)
        if isinstance(tc.k_folds, int) and tc.k_folds >= 2:
            self._num_folds = tc.k_folds
            models = list()
            for _ in range(tc.k_folds):
                models.append(Sequential())
            self._model = models
        else:
            self._num_folds = 1
            self._model = [Sequential()]
        self._train_history = [None] * self._num_folds
        # Arrays
        self._array_prepared = False
        self._model_meta = {
            "norm_dict": None,
        }
        # Report
        self._save_tb_logs = tc.save_tb_logs
        self._tb_logs_path = tc.tb_logs_path

    def get_hypers(self):
        """Gets a dictionary of hyper-parameters that defines a training"""
        tc = self._job_config.train.clone()
        hypers = {
            # hypers for building model
            "layers": tc.layers,
            "nodes": tc.nodes,
            "dropout_rate": tc.dropout_rate,
            "output_bkg_node_names": tc.output_bkg_node_names,
            "train_metrics": tc.train_metrics,
            "train_metrics_weighted": tc.train_metrics_weighted,
            "learn_rate": tc.learn_rate,
            "learn_rate_decay": tc.learn_rate_decay,
            "momentum": tc.momentum,
            "nesterov": tc.nesterov,
            # hypers for training
            "val_split": tc.val_split,
            "batch_size": tc.batch_size,
            "epochs": tc.epochs,
            "sig_class_weight": tc.sig_class_weight,
            "bkg_class_weight": tc.bkg_class_weight,
        }
        return hypers

    def get_inputs(
        self,
    ) -> Dict[str, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """Gets train/test datasets from feedbox"""
        ic = self._job_config.input.clone()
        ## get input
        input_df = self._feedbox.get_processed_df()
        cols = ic.selected_features
        # load train/test
        train_index = input_df["is_train"] == True
        test_index = input_df["is_train"] == False
        x_train = input_df.loc[train_index, cols].values
        x_test = input_df.loc[test_index, cols].values
        y_train = input_df.loc[train_index, ["y"]].values
        y_test = input_df.loc[test_index, ["y"]].values
        wt_train = input_df.loc[train_index, "weight"].values
        wt_test = input_df.loc[test_index, "weight"].values
        del input_df
        # remove negative weight events
        if ic.rm_negative_weight_events == True:
            wt_train = wt_train.clip(min=0)
            wt_test = wt_test.clip(min=0)

        return {
            "train": (x_train, y_train, wt_train),
            "test": (x_test, y_test, wt_test),
        }

    def get_train_callbacks(self, fold_num: Optional[int] = None) -> list:
        """Prepares callbacks of training"""
        train_callbacks = []
        tc = self._job_config.train.clone()
        if self._save_tb_logs:  # TODO: add back this function
            pass
            # if self._tb_logs_path is None:
            #    self._tb_logs_path = "temp_logs/{}".format(self._model_label)
            #    logger.warning(
            #        "TensorBoard logs path not specified, set path to: {}".format(
            #            self._tb_logs_path
            #        )
            #    )
            # tb_callback = TensorBoard(log_dir=self._tb_logs_path, histogram_freq=1)
            # train_callbacks.append(tb_callback)
        if tc.use_early_stop:
            early_stop_callback = callbacks.EarlyStopping(
                monitor=tc.early_stop_paras.monitor,
                min_delta=tc.early_stop_paras.min_delta,
                patience=tc.early_stop_paras.patience,
                mode=tc.early_stop_paras.mode,
                restore_best_weights=tc.early_stop_paras.restore_best_weights,
            )
            train_callbacks.append(early_stop_callback)
        ## set up check point to save model in each epoch
        if tc.save_model:
            model_save_dir = self.get_model_save_dir(fold_num=fold_num)
            pathlib.Path(model_save_dir).mkdir(parents=True, exist_ok=True)
            path_pattern = (
                f"{model_save_dir}/{self._model_name}_epoch{{epoch}}.h5"
            )
            checkpoint = ModelCheckpoint(path_pattern, monitor="val_loss")
            train_callbacks.append(checkpoint)
        return train_callbacks

    def train(self):
        # Check
        if not self._model_is_compiled:
            logger.critical("DNN model is not yet built, rebuilding")
            self.build()
        if not self._array_prepared:
            logger.critical("Training data is not ready, pleas set up inputs")
            exit(1)
        # Prepare
        tc = self._job_config.train.clone()
        input_dict = self.get_inputs()
        model = self.get_model()
        hypers = self.get_hypers()
        n_folds = self._num_folds
        verbose = tc.verbose
        # Input
        x_train, y_train, wt_train = input_dict["train"]
        x_test, y_test, wt_test = input_dict["test"]
        # Train
        logger.info("-" * 40)
        logger.info("Loading inputs")
        (
            train_index_list,
            validation_index_list,
        ) = train_utils.get_train_val_indices(
            y_train, y_train, wt_train, hypers["val_split"], k_folds=tc.k_folds
        )
        for fold_num in range(n_folds):
            logger.info(f"Training start. Using model: {self._model_name}")
            logger.info(f"Model info: {self._model_note}")
            if n_folds >= 2:
                logger.info(
                    f"Performing k-fold training {fold_num + 1}/{n_folds}"
                )
            fold_model = model[fold_num]
            fold_model.summary()
            train_index = train_index_list[fold_num]
            val_index = validation_index_list[fold_num]
            x_fold = x_train[train_index]
            y_fold = y_train[train_index]
            wt_fold = wt_train[train_index]
            val_x_fold = x_train[val_index]
            val_y_fold = y_train[val_index]
            val_wt_fold = wt_train[val_index]
            val_fold = (val_x_fold, val_y_fold, val_wt_fold)
            logger.info(
                f"> Training on {len(y_fold)}, validating on {len(val_y_fold)} events."
            )
            history_obj = fold_model.fit(
                x_fold,
                y_fold,
                batch_size=hypers["batch_size"],
                epochs=hypers["epochs"],
                validation_data=val_fold,
                shuffle=True,
                class_weight={
                    1: hypers["sig_class_weight"],
                    0: hypers["bkg_class_weight"],
                },
                sample_weight=wt_fold,
                callbacks=self.get_train_callbacks(fold_num=fold_num),
                verbose=verbose,
            )
            logger.info("Training finished.")
            # evaluation
            logger.info("Evaluate with test dataset:")
            score = fold_model.evaluate(
                x_test, y_test, verbose=verbose, sample_weight=wt_test,
            )
            if not isinstance(score, Iterable):
                logger.info(f"> test loss: {score}")
            else:
                for i, metric in enumerate(fold_model.metrics_names):
                    logger.info(f"> test - {metric}: {score[i]}")
            # save training details
            self._train_history[fold_num] = history_obj.history
            self.save_model(fold_num=fold_num)
            self.save_model_paras(fold_num=fold_num)
        # Update status
        self._model_is_trained = True


class Model_Sequential_Flat(Model_Sequential_Base):
    """Flat sequential model"""

    def __init__(self, job_config):
        super().__init__(job_config)

        self._model_label = "mod_seq"
        self._model_note = "Sequential model with flexible layers and nodes."
        self._tune_fun_name = "tune_Model_Sequential_Flat"

    def build(self):
        hypers = self.get_hypers()
        for fold_num in range(self._num_folds):
            fold_model = self._model[fold_num]
            self.build_single(fold_model, hypers)
        self._model_is_compiled = True

    def build_single(self, fold_model, hypers):
        # Add layers
        for layer in range(int(hypers["layers"])):
            # input layer
            if layer == 0:
                fold_model.add(
                    Dense(
                        hypers["nodes"],
                        kernel_initializer="glorot_uniform",
                        activation="relu",
                        input_dim=self._model_input_dim,
                    )
                )
            # hidden layers
            else:
                fold_model.add(
                    Dense(
                        hypers["nodes"],
                        kernel_initializer="glorot_uniform",
                        activation="relu",
                    )
                )
            if hypers["dropout_rate"] != 0:
                fold_model.add(Dropout(hypers["dropout_rate"]))
        # output layer
        if hypers["output_bkg_node_names"]:
            num_nodes_out = len(hypers["output_bkg_node_names"]) + 1
        else:
            num_nodes_out = 1
        fold_model.add(
            Dense(
                num_nodes_out,
                kernel_initializer="glorot_uniform",
                activation="sigmoid",
            )
        )
        # Compile
        # transfer self-defined metrics into real function
        metrics = copy.deepcopy(hypers["train_metrics"])
        weighted_metrics = copy.deepcopy(hypers["train_metrics_weighted"])
        if "plain_acc" in metrics:
            index = metrics.index("plain_acc")
            metrics[index] = plain_acc
        if "plain_acc" in weighted_metrics:
            index = weighted_metrics.index("plain_acc")
            weighted_metrics[index] = plain_acc
        if "auc" in weighted_metrics:
            index = weighted_metrics.index("auc")
            weighted_metrics[index] = tf.keras.metrics.AUC(name="auc")
        # compile model
        fold_model.compile(
            loss="binary_crossentropy",
            optimizer=SGD(
                lr=hypers["learn_rate"],
                decay=hypers["learn_rate_decay"],
                momentum=hypers["momentum"],
                nesterov=hypers["nesterov"],
            ),
            metrics=metrics,
            weighted_metrics=weighted_metrics,
        )

    def get_hypers_tune(self):
        """Gets a dict of hyperparameter (space) for auto-tuning"""
        ic = self._job_config.input.clone()
        rc = self._job_config.run.clone()
        model_cfg = self._job_config.tune.clone().model
        gs = train_utils.get_single_hyper
        hypers = {
            # hypers for building model
            "layers": gs(model_cfg.layers),
            "nodes": gs(model_cfg.nodes),
            "dropout_rate": gs(model_cfg.dropout_rate),
            "output_bkg_node_names": model_cfg.output_bkg_node_names,
            "tune_metrics": model_cfg.tune_metrics,
            "tune_metrics_weighted": model_cfg.tune_metrics_weighted,
            "learn_rate": gs(model_cfg.learn_rate),
            "learn_rate_decay": gs(model_cfg.learn_rate_decay),
            "momentum": gs(model_cfg.momentum),
            "nesterov": gs(model_cfg.nesterov),
            # hypers for training
            "batch_size": gs(model_cfg.batch_size),
            "epochs": gs(model_cfg.epochs),
            "sig_class_weight": gs(model_cfg.sig_class_weight),
            "bkg_class_weight": gs(model_cfg.bkg_class_weight),
            "use_early_stop": model_cfg.use_early_stop,
            "early_stop_paras": model_cfg.early_stop_paras.get_config_dict(),
            # job config
            "input_dim": len(ic.selected_features),
            # input dir
            "input_dir": str(pathlib.Path(rc.tune_input_cache).resolve()),
        }
        return hypers


# tuning functions for ray.tune

## Note: such functions take one config argument and report the score


def tune_Model_Sequential_Flat(config, checkpoint_dir=None):
    """Trainable function for ray-tune"""
    input_dir = pathlib.Path(config["input_dir"])
    x_train = np.load(input_dir / "x_train.npy")
    x_train_unreset = np.load(input_dir / "x_train_unreset.npy")
    y_train = np.load(input_dir / "y_train.npy")
    wt_train = np.load(input_dir / "wt_train.npy")
    x_val = np.load(input_dir / "x_val.npy")
    x_val_unreset = np.load(input_dir / "x_val_unreset.npy")
    y_val = np.load(input_dir / "y_val.npy")
    wt_val = np.load(input_dir / "wt_val.npy")

    # build model
    import tensorflow as tf  # must import inside function, other wise will get errors of ray tune

    model = tf.keras.models.Sequential()
    for layer in range(int(config["layers"])):
        ## input layer
        if layer == 0:
            model.add(
                tf.keras.layers.Dense(
                    config["nodes"],
                    kernel_initializer="glorot_uniform",
                    activation="relu",
                    input_dim=config["input_dim"],
                )
            )
        ## hidden layers
        else:
            model.add(
                tf.keras.layers.Dense(
                    config["nodes"],
                    kernel_initializer="glorot_uniform",
                    activation="relu",
                )
            )
        if config["dropout_rate"] != 0:
            model.add(tf.keras.layers.Dropout(config["dropout_rate"]))
    ## output layer
    if config["output_bkg_node_names"]:
        num_nodes_out = len(config["output_bkg_node_names"]) + 1
    else:
        num_nodes_out = 1
    model.add(
        tf.keras.layers.Dense(
            num_nodes_out,
            kernel_initializer="glorot_uniform",
            activation="sigmoid",
        )
    )
    metric_auc = tf.keras.metrics.AUC()
    model.compile(
        loss="binary_crossentropy",
        optimizer=tf.keras.optimizers.SGD(
            lr=config["learn_rate"],
            decay=config["learn_rate_decay"],
            momentum=config["momentum"],
            nesterov=config["nesterov"],
        ),
        metrics=config["tune_metrics"],
        weighted_metrics=list(config["tune_metrics_weighted"]) + [metric_auc],
    )

    # set callbacks
    callbacks = list()
    report_dict = {
        "auc": "auc",
        "val_auc": "val_auc",
    }  # default includes AUC of ROC
    for metric in config["tune_metrics"] + config["tune_metrics_weighted"]:
        report_dict[metric] = metric
        report_dict["val_" + metric] = "val_" + metric
    # tune_report_callback = TuneReportCallback(report_dict)
    # callbacks.append(tune_report_callback)
    if config["use_early_stop"]:
        es_config = config["early_stop_paras"]
        early_stop_callback = tf.keras.callbacks.EarlyStopping(**es_config)
        callbacks.append(early_stop_callback)

    # train
    # history_obj = model.fit(
    #    x_train,
    #    y_train,
    #    batch_size=config["batch_size"],
    #    epochs=config["epochs"],
    #    validation_data=(x_val, y_val, wt_val),
    #    shuffle=True,
    #    class_weight={
    #        1: config["sig_class_weight"],
    #        0: config["bkg_class_weight"],
    #    },
    #    sample_weight=wt_train,
    #    callbacks=callbacks,
    # )

    last_auc_unreset = 0
    for epoch_id in range(int(config["epochs"])):
        history_obj = model.fit(
            x_train,
            y_train,
            batch_size=config["batch_size"],
            epochs=1,
            validation_data=(x_val, y_val, wt_val),
            shuffle=True,
            class_weight={
                1: config["sig_class_weight"],
                0: config["bkg_class_weight"],
            },
            sample_weight=wt_train,
            callbacks=callbacks,
        )
        epoch_report = dict()
        for key in report_dict.keys():
            metric = str(key)
            epoch_report[metric] = history_obj.history[metric][-1]

        # y_pred = model.predict(x_val)
        # auc = roc_auc_score(y_val, y_pred, sample_weight=wt_val)
        # epoch_report["val_auc2"] = auc

        y_pred_unreset = model.predict(x_val_unreset)
        auc_unreset = roc_auc_score(
            y_val, y_pred_unreset, sample_weight=wt_val
        )
        epoch_report["auc_unreset"] = auc_unreset

        auc_unreset_improvement = auc_unreset - last_auc_unreset
        epoch_report["auc_unreset_improvement"] = auc_unreset_improvement
        last_auc_unreset = auc_unreset

        epoch_report["epoch_num"] = epoch_id + 1

        tune.report(**epoch_report)

    # evaluate metric
    # final_report = dict()
    # for key in report_dict.keys():
    #    metric = str(key)
    #    final_report[metric] = history_obj.history[metric][-1]
    # tune.report(**final_report)



import copy
import logging
import pathlib
import re

import yaml

from hepynet.common import common_utils
from hepynet.common.config_default import DEFAULT_CFG

logger = logging.getLogger("hepynet")


class Hepy_Config(object):
    """Helper class to handle job configs"""

    def __init__(self, config: dict) -> None:
        # define supported sections to avoid lint error
        self.config = Hepy_Config_Section({})
        self.job = Hepy_Config_Section({})
        self.input = Hepy_Config_Section({})
        self.train = Hepy_Config_Section({})
        self.tune = Hepy_Config_Section({})
        self.apply = Hepy_Config_Section({})
        self.para_scan = Hepy_Config_Section({})
        self.run = Hepy_Config_Section({})
        # set default
        self.update(DEFAULT_CFG)
        # initialize config
        for key, value in config.items():
            if type(value) is dict:
                getattr(self, key).update(value)
            elif value is None:
                pass
            else:
                logger.critical(
                    f"Expect section {key} must has dict type value or None, please check the input."
                )
                raise ValueError

    def clone(self):
        return copy.deepcopy(self)

    def get_config_dict(self) -> dict:
        """ Returns config in dict format """
        out_dict = {}
        out_dict["config"] = self.config.get_config_dict()
        out_dict["job"] = self.job.get_config_dict()
        out_dict["input"] = self.input.get_config_dict()
        out_dict["train"] = self.train.get_config_dict()
        out_dict["tune"] = self.tune.get_config_dict()
        out_dict["apply"] = self.apply.get_config_dict()
        out_dict["para_scan"] = self.para_scan.get_config_dict()
        out_dict["run"] = self.run.get_config_dict()
        return out_dict

    def update(self, config: dict) -> None:
        """Updates configs with given config dict, overwrite if exists

        Args:
            config (dict): two level dictionary of configs
        """
        for key, value in config.items():
            if type(value) is dict:
                if key in self.__dict__.keys():
                    getattr(self, key).update(value)
                else:
                    setattr(self, key, Hepy_Config_Section(value))
            elif value is None:
                pass
            else:
                logger.critical(
                    f"Expect section {key} has dict type value or None, please check the input."
                )
                raise ValueError

    def print(self) -> None:
        """Shows all configs
        """
        print("")
        print("Config details " + ">" * 80)
        for key, value in self.__dict__.items():
            print(f"[{key}]")
            value.print()
        print("Config ends " + "<" * 83)
        print("")


class Hepy_Config_Section(object):
    """Helper class to handle job configs in a section"""

    def __init__(self, section_config_dict: dict) -> None:
        for key, value in section_config_dict.items():
            if type(value) is dict:
                setattr(self, key, Hepy_Config_Section(value))
            else:
                setattr(self, key, value)

    def __deepcopy__(self, memo):
        clone_obj = Hepy_Config_Section(self.get_config_dict())
        return clone_obj

    def __getattr__(self, item):
        """Called when an attribute lookup has not found the attribute in the usual places"""
        return None

    def clone(self):
        return copy.deepcopy(self)

    def get_config_dict(self) -> dict:
        """ Returns config in dict format """
        config_dict = dict()
        for key, value in self.__dict__.items():
            if key != "_config_dict":
                if isinstance(value, Hepy_Config_Section):
                    config_dict[key] = value.get_config_dict()
                else:
                    config_dict[key] = value
        return config_dict

    def update(self, cfg_dict: dict) -> None:
        """Updates the section config dict with new dict, overwrite if exists

        Args:
            cfg_dict (dict): new section config dict for update
        """
        for key, value in cfg_dict.items():
            if type(value) is dict:
                if (
                    key in self.__dict__.keys()
                    and key != "_config_dict"
                    and isinstance(getattr(self, key), Hepy_Config_Section)
                ):
                    getattr(self, key).update(value)
                else:
                    setattr(self, key, Hepy_Config_Section(value))
            else:
                setattr(self, key, value)

    def print(self, tabs=0) -> None:
        """Shows all section configs
        """
        for key, value in self.__dict__.items():
            if key != "_config_dict":
                if isinstance(value, Hepy_Config_Section):
                    print(f"{' '*4*tabs}    {key} :")
                    value.print(tabs=tabs + 1)
                elif isinstance(value, list):
                    print(f"{' '*4*tabs}    {key} :")
                    for ele in value:
                        print(f"{' '*4*tabs}        - {ele}")
                else:
                    print(f"{' '*4*tabs}    {key} : {value}")


def load_current_platform_meta() -> dict:
    """Loads meta data for current platform

    Returns:
        dict: meta data dict of current platform
    """
    platform_meta = load_pc_meta()["platform_meta"]
    current_hostname = common_utils.get_current_hostname()
    current_platform = common_utils.get_current_platform_name()
    for my_host in platform_meta.keys():
        if re.match(f"({my_host})", current_hostname):
            for my_platform in platform_meta[my_host]:
                if re.match(f"({my_platform})", current_platform):
                    return platform_meta[my_host][my_platform]
    logger.warn(
        f"No meta data found for current host {current_hostname} with platform {current_platform}, please update the config at share/cross_platform/pc_meta.yaml"
    )
    return {"data_path": None}


def load_pc_meta() -> dict:
    """Loads platform meta dict for different path setup in different machines.

    Returns:
        dict: meta information dictionary
    """
    pc_meta_cfg_path = (
        f"{pathlib.Path().parent}/share/cross_platform/pc_meta.yaml"
    )
    try:
        with open(pc_meta_cfg_path) as pc_meta_file:
            pc_meta_dict = yaml.load(pc_meta_file, Loader=yaml.FullLoader)
            logger.debug(f"pc_meta config loaded: {pc_meta_cfg_path}")
            return pc_meta_dict
    except:
        logger.warn(
            "Can't load pc_meta config file, please check: share/cross_platform/pc_meta.yaml"
        )
        return {"platform_meta": {}}


def load_yaml_dict(yaml_path) -> dict:
    try:
        yaml_file = open(yaml_path, "r")
        return yaml.load(yaml_file, Loader=yaml.FullLoader)
    except:
        logger.critical(
            f"Can't read yaml config: {yaml_path}, please check whether input yaml config exists and the syntax is correct"
        )
        raise IOError



import argparse
import logging
import os
import time

logger = logging.getLogger("hepynet")
# Show tensorflow warnings and errors only
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"


def execute():
    """Enterpoint function of hepynet"""
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml_configs", nargs="*", action="store")
    parser.add_argument(
        "-d",
        "--debug",
        required=False,
        help="run in debug mode",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--resume",
        required=False,
        help="resume (tune) job",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--time",
        required=False,
        help="display time",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        required=False,
        help="verbose debug information",
        action="store_true",
    )
    args = parser.parse_args()

    # check input
    if not args.yaml_configs:
        parser.print_help()
        exit()
    else:
        # set logging level
        if args.debug:  # DEBUG, INFO, WARNING, ERROR, CRITICAL
            logging.getLogger("hepynet").setLevel(logging.DEBUG)
        else:
            logging.getLogger("hepynet").setLevel(logging.INFO)
        # set logging format
        logging_format = "%(levelname)s %(message)s"
        if args.verbose:
            logging_format += " @ %(filename)s:%(lineno)d - %(funcName)s"
        if args.time:
            logging_format = "%(asctime)s,%(msecs)03d " + logging_format
        logging.basicConfig(
            format=logging_format, datefmt="%Y-%m-%d:%H:%M:%S",
        )

        from hepynet.main import job_executor

        for yaml_cfg in args.yaml_configs:
            logger.info("#" * 80)
            logger.info(f"Executing: {yaml_cfg}")
            job_start_time = time.perf_counter()
            executor = job_executor.job_executor(yaml_cfg)
            executor.execute_jobs(resume=args.resume)
            job_end_time = time.perf_counter()
            time_consumed = job_end_time - job_start_time
            time_consumed_str = time.strftime(
                "%H:%M:%S", time.gmtime(time_consumed)
            )
            logger.info(f"Time consumed: {time_consumed_str}")
        logger.info("#" * 80)
        logger.info("Done!")
        logger.info("#" * 80)



"""Loads arrays for training.

Note:
  1. Arrays organized by dictionary, each key is corresponding to a sample component
  2. Each sample component is a sub-dictionary, each key is corresponding to a input feature's array 


TODO:
    Numpy input format has been replaced by pandas DataFrame.
    Should remove numpy_io or add support for different input formats?

"""
import logging
import pathlib
from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from hepynet.common import common_utils, config_utils
from hepynet.data_io import array_utils

logger = logging.getLogger("hepynet")


def get_array_key(job_config, sample_type):
    ic = job_config.input.clone()
    if sample_type == "sig":
        return ic.sig_key
    elif sample_type == "bkg":
        return ic.bkg_key
    elif sample_type == "data":
        return ic.data_key
    else:
        logger.error(f"Unknown sample type: {sample_type}")
        return None


def get_campaign_list(job_config):
    ic = job_config.input.clone()
    if "+" in ic.campaign:
        return [camp.strip() for camp in ic.campaign.split("+")]
    else:
        return [ic.campaign]


def get_cut_index(job_config, cut_feature_dict):
    ic = job_config.input.clone()
    cut_index = None
    for cut_feature, cut_value, cut_type in zip(
        ic.cut_features, ic.cut_values, ic.cut_types
    ):
        cut_array = cut_feature_dict[cut_feature]
        if cut_type == "=":
            temp_cut_index = np.argwhere(cut_array != cut_value)
        elif cut_type == ">":
            temp_cut_index = np.argwhere(cut_array <= cut_value)
        elif cut_type == "<":
            temp_cut_index = np.argwhere(cut_array >= cut_value)
        else:
            temp_cut_index = np.array([])
        # update cut_index
        if cut_index is None:
            cut_index = temp_cut_index
        else:
            cut_index = np.union1d(cut_index, temp_cut_index)
    return cut_index


def get_cut_pass_index_dict(
    job_config, array_type_list=None, sample_list_dict=None
) -> Dict[str, Dict[str, np.ndarray]]:
    job_config_copy = job_config.clone()
    ic = job_config_copy.input
    # rc = job_config_copy.run
    # ic.cut_features += [ic.channel]
    # ic.cut_values += [1]
    # ic.cut_types += ["="]
    pass_id_dict = dict()
    if array_type_list is None:
        array_type_list = ["sig", "bkg", "data"]
    for array_type in array_type_list:
        pass_id_dict[array_type] = dict()
        sample_load_list = list()
        if (sample_list_dict) is None or (array_type not in sample_list_dict):
            samples = getattr(ic, f"{array_type}_list")
            if type(samples) != list:
                sample_load_list = [samples]
            else:
                sample_load_list = samples
        else:
            sample_load_list = sample_list_dict[array_type]
        data_dir = get_data_dir()
        ## get campaign list
        campaign_list = list()
        if "+" in ic.campaign:
            campaign_list = [camp.strip() for camp in ic.campaign.split("+")]
        else:
            campaign_list = [ic.campaign]
        ## load sample array
        for sample_component in sample_load_list:
            cut_array_dict = {}
            for feature in ic.cut_features:
                feature_array = None
                for camp in campaign_list:
                    temp_array = np.load(
                        f"{data_dir}/{ic.arr_path}/{ic.arr_version}/{ic.variation}/{ic.channel}/{camp}/{ic.region}/{sample_component}_{feature}.npy ",
                    )
                    if feature_array is None:
                        feature_array = temp_array
                    else:
                        feature_array = np.concatenate(
                            (feature_array, temp_array)
                        )

                feature_array = feature_array

                cut_array_dict[feature] = feature_array.flatten()
            # apply cuts
            ## Get indexes that pass cuts
            if not (
                len(ic.cut_features) == len(ic.cut_values)
                and len(ic.cut_features) == len(ic.cut_types)
            ):
                logger.critical(
                    "cut_features and cut_values and cut_types should have same length"
                )
                exit(1)
            pass_index = None
            for cut_feature, cut_value, cut_type in zip(
                ic.cut_features, ic.cut_values, ic.cut_types
            ):
                temp_pass_index = array_utils.get_cut_index_value(
                    cut_array_dict[cut_feature], cut_value, cut_type
                )
                if pass_index is None:
                    pass_index = temp_pass_index
                else:
                    pass_index = np.intersect1d(pass_index, temp_pass_index)
            logger.debug(
                f"Got cut-passed index for {array_type} {sample_component}"
            )
            if pass_index is not None:
                logger.debug(f"Shape {pass_index.shape}")
            pass_id_dict[array_type][sample_component] = pass_index
    return pass_id_dict


def get_data_dir():
    platform_meta = config_utils.load_current_platform_meta()
    data_dir = platform_meta["data_path"]
    ## check data_root_directory
    if not data_dir:
        current_hostname = common_utils.get_current_hostname()
        current_platform = common_utils.get_current_platform_name()
        logger.info(
            f"NO data_path setting for current host {current_hostname} with platform {current_platform} found at share/cross_platform/pc_meta.yaml, will consider using absolute path for input data"
        )
        data_dir = ""
    return data_dir


def get_samples_total_weight(job_config, sample_list):
    ic = job_config.input.clone()
    data_dir = get_data_dir()
    total_weight = 0
    for sample in sample_list:
        for camp in get_campaign_list(job_config):
            npy_in_dir = pathlib.Path(
                f"{data_dir}/{ic.arr_path}/{ic.arr_version}/{ic.variation}/{ic.channel}/{camp}/{ic.region}"
            )
            wt_path = npy_in_dir / f"{sample}_weight.npy"
            wt = np.load(wt_path)
            total_weight += np.sum(wt)
    return total_weight


def get_sumofweight(job_config, sample_type):
    ic = job_config.input.clone()
    if sample_type == "sig":
        return ic.sig_sumofweight
    elif sample_type == "bkg":
        return ic.bkg_sumofweight
    elif sample_type == "data":
        return ic.data_sumofweight
    else:
        logger.error(f"Unknown sample type: {sample_type}")
        return None


def load_npy_arrays(
    job_config,
    array_type,
    part_features: Optional[List[str]] = None,
    cut_pass_index_dict=None,
    include_weight=True,
):
    """Gets individual npy arrays with given info.

    Return:
        A dict of subdict for different samples
            example: signal dict of rpv_500 rpv_1000 rpv_2000
        Each subdict is a dict of different variable
            examples: pT, eta, phi

    """
    # get settings from config file
    job_config_copy = job_config.clone()
    rc = job_config_copy.run
    ic = job_config_copy.input
    if array_type == "sig":
        samples = ic.sig_list
    elif array_type == "bkg":
        samples = ic.bkg_list
    elif array_type == "data":
        samples = ic.data_list
    else:
        logger.critical(f"Unsupported array_type: {array_type}")
        exit(1)

    # check different situation
    if type(samples) != list:
        sample_list = [samples]
    else:
        sample_list = samples

    # load arrays
    out_features = list()
    if part_features is None:
        out_features += ic.selected_features
        out_features += ic.validation_features
    else:
        out_features += part_features
    if include_weight:
        out_features += ["weight"]
    out_features = list(set().union(out_features))  # remove duplicates
    logger.debug(f"Out_features {out_features}")
    out_dict = {}
    ## check data_root_directory
    data_dir = get_data_dir()
    ## get campaign list
    campaign_list = list()
    if "+" in ic.campaign:
        campaign_list = [camp.strip() for camp in ic.campaign.split("+")]
    else:
        campaign_list = [ic.campaign]
    ## load sample array
    for sample_component in sample_list:
        pass_index = None
        if (
            isinstance(cut_pass_index_dict, dict)
            and array_type in cut_pass_index_dict
            and sample_component in cut_pass_index_dict[array_type]
        ):
            pass_index = cut_pass_index_dict[array_type][sample_component]
        else:
            feature_cut_dict = get_cut_pass_index_dict(
                job_config,
                array_type_list=[array_type],
                sample_list_dict={array_type: [sample_component]},
            )
            pass_index = feature_cut_dict[array_type][sample_component]
        sample_array_dict = {}
        for feature in out_features:
            feature_array = None
            for camp in campaign_list:
                load_path = f"{data_dir}/{ic.arr_path}/{ic.arr_version}/{ic.variation}/{ic.channel}/{camp}/{ic.region}/{sample_component}_{feature}.npy "
                temp_array = np.load(load_path)
                if feature_array is None:
                    feature_array = temp_array
                else:
                    feature_array = np.concatenate((feature_array, temp_array))
            # use float32 to reduce memory consumption, risky?
            feature_array = np.array(feature_array, dtype=np.float32)
            if pass_index is not None:
                sample_array_dict[feature] = feature_array[
                    pass_index
                ].flatten()
            else:
                sample_array_dict[feature] = feature_array.flatten()
        if "weight" in out_features:
            weight_array = sample_array_dict["weight"]
            total_weights = np.sum(weight_array)
            logger.debug(
                f"Load sample {sample_component} total weights: {total_weights}, total events {weight_array.shape}"
            )
        out_dict[sample_component] = sample_array_dict
    return out_dict


def save_npy_array(
    npy_array, save_path,
):
    save_path = pathlib.Path(save_path)
    save_path.cwd().mkdir(parents=True, exist_ok=True)
    with open(save_path, "wb") as f:
        np.save(f, npy_array)
    logger.debug(f"Array saved to: {save_path}")
    logger.debug(f"Array contents:\n {npy_array}")
    logger.debug(f"Array average:\n {np.average(npy_array)}")
    logger.debug(f"Array sum:\n {np.sum(npy_array)}")



import logging

import numpy as np
import pandas as pd

logger = logging.getLogger("hepynet")


def cov(x, y, w):
    """Calculates weighted covariance"""
    return np.sum(
        w
        * (x - np.average(x, axis=0, weights=w))
        * (y - np.average(y, axis=0, weights=w))
    ) / np.sum(w)


def corr(x, y, w):
    """Calculates weighted correlation"""
    return cov(x, y, w) / np.sqrt(cov(x, x, w) * cov(y, y, w))


def corr_matrix(x, w=None):
    """Calculates correlation coefficient matrix"""
    if w is None:
        w = np.ones(x.shape[0])
    num_features = x.shape[1]
    # np.cov can't deal with negative weights, use self-defined function for now
    corr_m = np.zeros((num_features, num_features))
    for row in range(num_features):  # TODO: need to optimize the algorithm
        for col in range(row + 1):
            corr_m[row][col] = corr(x[:, row], x[:, col], w)
            corr_m[col][row] = corr_m[row][col]
    return corr_m


def extract_bkg_df(df: pd.DataFrame):
    return df.loc[(df["is_mc"] == True) & (df["is_sig"] == False)]


def extract_sig_df(df: pd.DataFrame):
    return df.loc[df["is_sig"] == True]


def get_cut_index(np_array, cut_values, cut_types):
    """Parses cuts arguments and returns cuts indexes."""
    assert len(cut_values) == len(
        cut_types
    ), "cut_values and cut_types should have same length."
    pass_index = None
    for cut_value, cut_type in zip(cut_values, cut_types):
        temp_index = get_cut_index_value(np_array, cut_value, cut_type)
        if pass_index is None:
            pass_index = temp_index
        else:
            pass_index = np.intersect1d(pass_index, temp_index)
    return pass_index


def get_cut_index_value(np_array: np.ndarray, cut_value, cut_type):
    """Returns cut indexes based on cut_value and cut_type.

    If cut_type is "=":
        returns all indexes that have values equal to cut_value
    If cut_type is ">":
        returns all indexes that have values lager than cut_value
    If cut_type is "<":
        returns all indexes that have values smaller than cut_value

    Args:
        array_dict: numpy array
        cut_feature: str
        cut_bool: bool
    """
    logger.debug("@ data_io.array_utils.get_cut_index_value")
    assert np_array.ndim == 1
    logger.debug(f"Cut input shape: {np_array.shape}, cut value {cut_value}")
    # Make cuts
    if cut_type == "=":
        pass_index = np.argwhere(np_array == cut_value)
    elif cut_type == ">":
        pass_index = np.argwhere(np_array > cut_value)
    elif cut_type == "<":
        pass_index = np.argwhere(np_array < cut_value)
    else:
        raise ValueError("Invalid cut_type specified.")
    return pass_index.flatten()



import logging
import pathlib
from typing import List

import atlas_mpl_style as ampl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.ticker import AutoMinorLocator, LogLocator

import hepynet.common.hepy_type as ht
from hepynet.data_io import array_utils

logger = logging.getLogger("hepynet")


def plot_correlation_matrix(
    df: pd.DataFrame, job_config: ht.config, save_dir: ht.pathlike = "."
):
    ic = job_config.input.clone()
    features = ic.selected_features
    save_dir = pathlib.Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)
    fig_scale = len(features) / 10 + 1
    figsize = (8.333 * fig_scale, 4.167 * fig_scale)
    fig, ax = plt.subplots(ncols=2, figsize=figsize)
    # plot bkg
    ax[0].set_title("bkg correlation")
    bkg_df = array_utils.extract_bkg_df(df)
    if bkg_df.shape[0] > 1000000:
        logger.warn(
            f"Too large input detected ({bkg_df.shape[0]} rows), randomly sampling 1000000 rows for background corr_matrix calculation"
        )
        bkg_df = bkg_df.sample(n=1000000)
    bkg_matrix = array_utils.corr_matrix(
        bkg_df[features].to_numpy("float32"),
        bkg_df["weight"].to_numpy("float32"),
    )
    paint_correlation_matrix(ax[0], bkg_matrix, features)
    # plot sig
    ax[1].set_title("sig correlation")
    sig_df = array_utils.extract_sig_df(df)
    if sig_df.shape[0] > 1000000:
        logger.warn(
            f"Too large input detected ({sig_df.shape[0]} rows), randomly sampling 1000000 rows for signal corr_matrix calculation"
        )
        sig_df = sig_df.sample(n=1000000)
    sig_matrix = array_utils.corr_matrix(
        sig_df[features].to_numpy("float32"),
        sig_df["weight"].to_numpy("float32"),
    )
    paint_correlation_matrix(ax[1], sig_matrix, features)
    fig_save_path = save_dir / "correlation_matrix.png"
    logger.debug(f"Save correlation matrix to: {fig_save_path}")
    fig.savefig(fig_save_path)


def paint_correlation_matrix(
    ax: ht.ax, corr_matrix: np.ndarray, labels: List[str]
):
    # Generate a mask for the upper triangle
    mask = np.triu(np.ones_like(corr_matrix, dtype=np.bool))
    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(220, 10, as_cmap=True)
    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(
        corr_matrix,
        mask=mask,
        cmap=cmap,
        vmax=0.3,
        center=0,
        square=True,
        xticklabels=labels,
        yticklabels=labels,
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
        ax=ax,
    )


def plot_input(
    df: pd.DataFrame,
    job_config: ht.config,
    save_dir: ht.pathlike = None,
    is_raw: bool = True,
):
    """Plots input distributions comparision plots for sig/bkg/data"""
    # setup config
    ic = job_config.input.clone()
    ac = job_config.apply.clone()
    plot_cfg = ac.cfg_kine
    # prepare
    plot_feature_list = list(
        set(ic.selected_features + ic.validation_features)
    )
    save_dir = pathlib.Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)
    # plot
    for feature in plot_feature_list:
        feature_cfg = plot_cfg.clone()
        if feature in plot_cfg.__dict__.keys():
            feature_cfg_tmp = getattr(plot_cfg, feature)
            feature_cfg.update(feature_cfg_tmp.get_config_dict())
        if is_raw:
            plot_range = feature_cfg.range_raw
            bkg_scale = feature_cfg.bkg_scale_raw
            sig_scale = feature_cfg.sig_scale_raw
        else:
            plot_range = feature_cfg.range_processed
            bkg_scale = feature_cfg.bkg_scale_processed
            sig_scale = feature_cfg.sig_scale_processed
        if feature_cfg.logbin and plot_range:
            plot_bins = np.logspace(
                np.log10(plot_range[0]),
                np.log10(plot_range[1]),
                feature_cfg.bins,
            )
        else:
            plot_bins = feature_cfg.bins
        # plot bkg
        # bkg_df = array_utils.extract_bkg_df(df)
        bkg_wt = df.loc[
            (df["is_mc"] == True) & (df["is_sig"] == False), "weight"
        ].to_numpy("float32")
        fig, ax = plt.subplots()
        hist_kwargs = feature_cfg.hist_kwargs_bkg.get_config_dict()
        remove_hist_kwargs_duplicates(hist_kwargs)
        ax.hist(
            df.loc[
                (df["is_mc"] == True) & (df["is_sig"] == False), feature
            ].to_numpy("float32"),
            bins=plot_bins,
            range=plot_range,
            weights=bkg_wt * bkg_scale,
            label="background",
            **(hist_kwargs),
        )
        modify_hist(ax, feature_cfg, plot_range)
        ax.set_xlabel(feature)
        # decide wether plot in same place
        if plot_cfg.separate_bkg_sig:
            if ac.plot_atlas_label:
                ampl.plot.draw_atlas_label(
                    0.05, 0.95, ax=ax, **(ac.atlas_label.get_config_dict())
                )
            # fig.suptitle(feature)
            fig.savefig(f"{save_dir}/{feature}_bkg.{plot_cfg.save_format}")
            plt.close()
            fig, ax = plt.subplots()
        # plot sig
        sig_df = array_utils.extract_sig_df(df)
        sig_wt = sig_df["weight"].to_numpy("float32")
        hist_kwargs = feature_cfg.hist_kwargs_sig.get_config_dict()
        remove_hist_kwargs_duplicates(hist_kwargs)
        ax.hist(
            sig_df[feature].to_numpy("float32"),
            bins=plot_bins,
            range=plot_range,
            weights=sig_wt * sig_scale,
            label="signal",
            **(hist_kwargs),
        )
        modify_hist(ax, feature_cfg, plot_range)
        ax.set_xlabel(feature)
        if ac.plot_atlas_label:
            ampl.plot.draw_atlas_label(
                0.05, 0.95, ax=ax, **(ac.atlas_label.get_config_dict())
            )
        # fig.suptitle(feature)
        if plot_cfg.separate_bkg_sig:
            fig.savefig(f"{save_dir}/{feature}_sig.{plot_cfg.save_format}")
        else:
            fig.savefig(f"{save_dir}/{feature}.{plot_cfg.save_format}")
        plt.close()


def plot_input_dnn(
    df_raw: pd.DataFrame,
    df: pd.DataFrame,
    job_config: ht.config,
    dnn_cut: float = None,
    multi_class_cut_branch: int = 0,
    save_dir: ht.pathlike = None,
):
    """Plots input distributions comparision plots with DNN cuts applied"""
    logger.info(
        f"Plotting input distributions with DNN cuts {dnn_cut} applied."
    )
    # prepare
    ic = job_config.input.clone()
    ac = job_config.apply.clone()
    plot_cfg = ac.cfg_cut_kine_study
    # get sig/bkg DataFrame and weights
    bkg_df_raw = array_utils.extract_bkg_df(df_raw)
    sig_df_raw = array_utils.extract_sig_df(df_raw)
    bkg_weights = bkg_df_raw["weight"].to_numpy("float32")
    sig_weights = sig_df_raw["weight"].to_numpy("float32")
    # get predictions
    bkg_predictions = array_utils.extract_bkg_df(df)[["y_pred"]].to_numpy(
        "float32"
    )
    sig_predictions = array_utils.extract_sig_df(df)[["y_pred"]].to_numpy(
        "float32"
    )
    # normalize
    if plot_cfg.density:
        bkg_weights = bkg_weights / np.sum(bkg_weights)
        sig_weights = sig_weights / np.sum(sig_weights)
    # plot kinematics with dnn cuts
    if dnn_cut < 0 or dnn_cut > 1:
        logger.error(f"DNN cut {dnn_cut} is out of range [0, 1]!")
        return
    # prepare signal
    sig_predictions = sig_predictions[:, multi_class_cut_branch]
    sig_cut_index = array_utils.get_cut_index(
        sig_predictions, [dnn_cut], ["<"]
    )
    sig_weights_dnn = sig_weights.copy()
    sig_weights_dnn[sig_cut_index] = 0
    # prepare background
    bkg_predictions = bkg_predictions[:, multi_class_cut_branch]
    bkg_cut_index = array_utils.get_cut_index(
        bkg_predictions, [dnn_cut], ["<"]
    )
    bkg_weights_dnn = bkg_weights.copy()
    bkg_weights_dnn[bkg_cut_index] = 0
    # normalize weights for density plots
    if plot_cfg.density:
        bkg_weights_dnn = bkg_weights_dnn / np.sum(bkg_weights)
        sig_weights_dnn = sig_weights_dnn / np.sum(sig_weights)
    # plot
    plot_feature_list = ic.selected_features + ic.validation_features
    for feature in plot_feature_list:
        logger.debug(f"Plotting kinematics for {feature}")
        bkg_array = bkg_df_raw[feature].to_numpy("float32")
        sig_array = sig_df_raw[feature].to_numpy("float32")

        feature_cfg = plot_cfg.clone()
        if feature in plot_cfg.__dict__.keys():
            feature_cfg_tmp = getattr(plot_cfg, feature)
            feature_cfg.update(feature_cfg_tmp.get_config_dict())
        plot_range = feature_cfg.range_processed

        # plot sig
        fig, main_ax, ratio_ax = ampl.ratio_axes()
        main_ax.hist(
            sig_array,
            bins=feature_cfg.bins,
            range=plot_range,
            weights=sig_weights,
            histtype="step",
            color=feature_cfg.sig_color,
            label="before DNN cut",
        )
        main_ax.hist(
            sig_array,
            bins=feature_cfg.bins,
            range=plot_range,
            weights=sig_weights_dnn,
            histtype="stepfilled",
            color=feature_cfg.sig_color,
            label="after DNN cut",
        )
        sig_bins, sig_edges = np.histogram(
            sig_array, bins=feature_cfg.bins, weights=sig_weights,
        )
        sig_bins_dnn, _ = np.histogram(
            sig_array, bins=feature_cfg.bins, weights=sig_weights_dnn,
        )
        ampl.plot.plot_ratio(
            sig_edges,
            sig_bins_dnn,
            np.zeros(len(sig_bins)),
            sig_bins,
            np.zeros(len(sig_bins)),
            ratio_ax,
            plottype="raw",
        )
        if plot_range:
            main_ax.set_xlim(plot_range[0], plot_range[1])
            ratio_ax.set_xlim(plot_range[0], plot_range[1])
        if feature_cfg.log:
            main_ax.set_yscale("log")
            _, y_max = main_ax.get_ylim()
            main_ax.set_ylim(
                feature_cfg.logy_min, y_max * np.power(10, np.log10(y_max) / 2)
            )
        else:
            _, y_max = main_ax.get_ylim()
            main_ax.set_ylim(0, y_max * 1.4)
        main_ax.legend(loc="upper right")
        # main_ax.set_ylabel(feature_cfg.y_label)
        ratio_ax.set_xlabel(feature)
        if ac.plot_atlas_label:
            ampl.plot.draw_atlas_label(
                0.05, 0.95, ax=main_ax, **(ac.atlas_label.get_config_dict())
            )
        fig.suptitle(f"{feature} (signal)")
        fig.savefig(f"{save_dir}/{feature}(sig).{plot_cfg.save_format}")
        plt.close()
        # plot bkg
        fig, main_ax, ratio_ax = ampl.ratio_axes()
        main_ax.hist(
            bkg_array,
            bins=feature_cfg.bins,
            range=plot_range,
            weights=bkg_weights,
            histtype="step",
            color=feature_cfg.bkg_color,
            label="before DNN cut",
        )
        main_ax.hist(
            bkg_array,
            bins=feature_cfg.bins,
            range=plot_range,
            weights=bkg_weights_dnn,
            histtype="stepfilled",
            color=feature_cfg.bkg_color,
            label="after DNN cut",
        )
        bkg_bins, bkg_edges = np.histogram(
            bkg_array, bins=feature_cfg.bins, weights=bkg_weights,
        )
        bkg_bins_dnn, _ = np.histogram(
            bkg_array, bins=feature_cfg.bins, weights=bkg_weights_dnn,
        )
        ampl.plot.plot_ratio(
            bkg_edges,
            bkg_bins_dnn,
            np.zeros(len(bkg_bins)),
            bkg_bins,
            np.zeros(len(bkg_bins)),
            ratio_ax,
            plottype="raw",
        )
        if plot_range:
            main_ax.set_xlim(plot_range[0], plot_range[1])
            ratio_ax.set_xlim(plot_range[0], plot_range[1])
        if feature_cfg.log:
            main_ax.set_yscale("log")
            _, y_max = main_ax.get_ylim()
            main_ax.set_ylim(
                feature_cfg.logy_min, y_max * np.power(10, np.log10(y_max) / 2)
            )
        else:
            _, y_max = main_ax.get_ylim()
            main_ax.set_ylim(0, y_max * 1.4)
        main_ax.legend(loc="upper right")
        # main_ax.set_ylabel(feature_cfg.y_label)
        ratio_ax.set_xlabel(feature)
        if ac.plot_atlas_label:
            ampl.plot.draw_atlas_label(
                0.05, 0.95, ax=main_ax, **(ac.atlas_label.get_config_dict())
            )
        fig.suptitle(f"{feature} (background)")
        fig.savefig(f"{save_dir}/{feature}(bkg).{plot_cfg.save_format}")
        plt.close()


def remove_hist_kwargs_duplicates(hist_kwargs: dict):
    if "bins" in hist_kwargs:
        del hist_kwargs["bins"]
    if "logbin" in hist_kwargs:
        del hist_kwargs["logbin"]
    if "logx" in hist_kwargs:
        del hist_kwargs["logx"]
    if "logy" in hist_kwargs:
        del hist_kwargs["logy"]
    if "range" in hist_kwargs:
        del hist_kwargs["range"]
    if "weights" in hist_kwargs:
        del hist_kwargs["weights"]
    if "label" in hist_kwargs:
        del hist_kwargs["label"]
    return hist_kwargs


def modify_hist(ax: ht.ax, feature_cfg: ht.sub_config, plot_range: ht.bound):
    ax.set_xlabel(feature_cfg.x_label)
    ax.set_ylabel(feature_cfg.y_label)
    log_minor_locator = LogLocator(
        base=10, subs=np.arange(1.0, 10.0) * 0.1, numticks=10
    )
    if feature_cfg.logx:
        ax.set_xscale("symlog")
        ax.xaxis.set_minor_locator(log_minor_locator)
    if plot_range:
        ax.set_xlim(plot_range[0], plot_range[1])
    y_min, y_max = ax.get_ylim()
    if feature_cfg.logy:
        ax.set_yscale("symlog")
        log_range = np.log10(y_max)
        y_max = pow(10, np.log10(y_max) + log_range * 0.4)
        ax.set_ylim(y_min, y_max * 1.4)
        ax.yaxis.set_minor_locator(log_minor_locator)
    else:
        ax.set_ylim(y_min, y_max * 1.4)
        ax.yaxis.set_minor_locator(AutoMinorLocator())
    ax.legend(loc="upper right")



DEFAULT_CFG = {
    "config": {
        "include": [],
        "best_tune_overwrite": False,  # overwrite with best tuned config
    },
    "job": {
        "job_name": "JOB_NAME_DEF",
        "job_type": "JOB_TYPE_DEF",  # could be train/tune/apply
        "save_dir": ".",
        "tune_job_name": "TUNE_JOB_NAME_DEF",
        "train_job_name": "TRAIN_JOB_NAME_DEF",
        "fix_rdm_seed": True,
        "rdm_seed": 1024,
        "date_str": None,
        "fix_date_str": False,
    },
    "input": {
        "input_path": "",
        "reshape_input": True,
        "reweight_input": True,
        "negative_weight_process": "to_zero",  # could also be "to_positive" or "keep"
        "feature_norm_alias": {},
        "sig_sumofweight": 1000,
        "bkg_sumofweight": 1000,
        "data_sumofweight": 1000,
        "reset_feature": False,
        "reset_feature_name": "RESET_FEATURE_NAME_DEF",
        "reset_feature_overwrite": False,
        "reset_feature_overwrite_value": None,
        "rm_negative_weight_events": False,
        "arr_path": "",
        "arr_version": "",
        "variation": "",
        "channel": "",
        "campaign": "",
        "region": "",
        "channel": "",
        "sig_key": "all",
        "sig_list": [],
        "bkg_key": "all",
        "bkg_list": [],
        "data_key": "all",
        "data_list": [],
        "selected_features": [],
        "validation_features": [],
        "cut_features": [],
        "cut_values": [],
        "cut_types": [],
        "cut_expression": None,
        "test_rate": 0.2,
    },
    "tune": {
        "tuner": {
            "init": {
                "num_cpus": None,
                "num_gpus": None,
                "log_to_driver": False,
            },
            "scheduler_class": "AsyncHyperBandScheduler",
            "scheduler": {
                "time_attr": "training_iteration",
                "max_t": 200,
                "grace_period": 10,
            },
            "algo_class": None,
            "algo": {},
            "stopper_class": None,
            "stopper": {},
            "run": {
                "num_samples": 10,
                "resources_per_trial": {"cpu": 1, "gpu": 0},
                "log_to_file": True,
            },
        },
        "model_class": "Model_Sequential_Flat",
        "model": {
            "output_bkg_node_names": [],
            "layers": 1,
            "nodes": 1,
            "learn_rate": 0.01,
            "learn_rate_decay": 0.01,
            "batch_size": 32,
            "epochs": 1,
            "momentum": 0,
            "nesterov": False,
            "dropout_rate": 0,
            "sig_class_weight": 1,
            "bkg_class_weight": 1,
            "val_split": 0.25,
            "use_early_stop": False,
            "early_stop_paras": {
                "monitor": "val_loss",
                "min_delta": 0,
                "patience": 1,
                "mode": "min",
                "restore_best_weights": True,
            },
            "tune_metrics": [],
            "tune_metrics_weighted": [],
        },
        "tmp_dir": "run/tmp_log",
        "rm_tmp_log": False,
    },
    "train": {
        "model_name": "MODEL_NAME_DEF",
        "model_class": "Model_Sequential_Flat",
        "output_bkg_node_names": [],
        "layers": 1,
        "nodes": 1,
        "learn_rate": 0.01,
        "learn_rate_decay": 0.01,
        "batch_size": 32,
        "epochs": 1,
        "momentum": 0,
        "nesterov": False,
        "dropout_rate": 0,
        "sig_class_weight": 1,
        "bkg_class_weight": 1,
        "val_split": 0.25,
        "k_folds": None,
        "use_early_stop": False,
        "early_stop_paras": {
            "monitor": "val_loss",
            "min_delta": 0,
            "patience": 1,
            "mode": "min",
            "restore_best_weights": True,
        },
        "train_metrics": [],
        "train_metrics_weighted": ["accuracy"],
        "save_model": True,
        "verbose": 0,
    },
    "apply": {
        "color_cycle": [
            "#f99157",  # orange
            "#5fb3b3",  # cyan
            "#ec5f67",  # red
            "#6699cc",  # blue
            "#fac863",  # yellow
            "#99c794",  # green
            "#c594c5",  # pink
            "#ab7967",  # brown
            "#cdd3de",  # grey
            "#343d46",  # black
        ],
        "plot_atlas_label": False,
        "atlas_label": {},
        "check_model_epoch": False,
        "epoch_check_interval": 5,
        "book_history": False,
        "cfg_history": {},
        "book_kine": False,
        "cfg_kine": {
            "separate_bkg_sig": True,
            "bkg_scale_raw": 1,
            "bkg_scale_processed": 1,
            "sig_scale_raw": 1,
            "sig_scale_processed": 1,
            "hist_kwargs_sig": {  # kwargs of plt.hist
                "bins": 40,
                "logbin": False,
                "logx": False,
                "logy": False,
                "histtype": "step",
                "density": False,
                "facecolor": "#f99157",  # orange
                "edgecolor": "#f99157",
                "lw": 2,
            },
            "hist_kwargs_bkg": {
                "bins": 40,
                "logbin": False,
                "logx": False,
                "logy": False,
                "histtype": "stepfilled",
                "density": False,
                "facecolor": "#5fb3b3",  # cyan
                "edgecolor": "#5fb3b3",
            },
            "save_format": "png",
            "x_label": "",
            "y_label": "",
            # dedicated feature configs can be made like the following:
            ## feature:
            ##   bins: 25
            ##   range: [10, 200]
            # feature config will overwrite common plot configs
        },
        "book_cor_matrix": False,
        # model dependent studies
        "jump_model_studies": False,
        "book_fit_inputs": False,
        "fit_df": {"region": "", "branches": [], "save_dir": "",},
        "book_confusion_matrix": False,
        "cfg_confusion_matrix": {"dnn_cut": 0.5},
        "book_roc": False,
        "book_pr": False,
        "book_train_test_compare": False,
        "cfg_train_test_compare": {
            "sig_key": None,
            "bkg_key": None,
            "plot_title": "train/test MVA scores compare",
            "bkg_color": "#f99157",  # orange
            "sig_color": "#5fb3b3",  # cyan
            "bins": 25,
            "range": [0, 1],
            "density": False,
            "log": False,
            "logy_min": 0.001,
            "save_format": "png",
        },
        "book_mva_scores_data_mc": False,
        "cfg_mva_scores_data_mc": {
            "sig_list": [],
            "bkg_list": [],
            "apply_data": False,
            "show_ratio": False,
            "data_key": "",
            "data_range": None,
            "plot_title": "MVA scores",
            "sig_color": "#5fb3b3",  # cyan
            "bkg_color": "#f99157",  # orange
            "sig_scale": 1.0,
            "bkg_scale": 1.0,
            "data_scale": 1.0,
            "bins": 25,
            "range": [0, 1],
            "density": False,
            "log": False,
            "logy_min": 0.001,
            "save_format": "png",
        },
        "book_significance_scan": False,
        "cfg_significance_scan": {"significance_algo": "s_sqrt_b_rel",},
        "book_cut_kine_study": False,
        "cfg_cut_kine_study": {
            "separate_sig_bkg": False,
            "save_ratio_table": False,
            "bins": 40,
            "range": None,
            "histtype": "stepfilled",
            "alpha": 0.3,
            "density": False,
            "bkg_color": "#f99157",  # orange
            "sig_color": "#5fb3b3",  # cyan
            "dnn_cut_list": None,
            "save_format": "png",
        },
        "book_importance_study": False,
        "cfg_importance_study": {"log": False},
    },
    "run": {"datestr": "", "config_collected": False, "input_dim": None,},
}



import csv
import logging
import pathlib
from typing import Iterable

import git

import hepynet.common.hepy_type as ht

logger = logging.getLogger("hepynet")

MAIN_DIR_NAMES = ["pdnn-lfv", "work"]


def get_valid_cfg_path(path: ht.pathlike):
    """Finds valid path for cfg file in /share folder.

    If path is already valid:
      Nothing will be done and original path will be returned.
    If path is not valid:
      Try to add share folder path before to see whether we can get a valid path.
      Otherwise, raise error to ask configuration correction.

    """
    # Check path:
    if pathlib.Path(path).is_file():
        return path
    # Check try add share folder prefix
    repo = git.Repo(".", search_parent_directories=True)
    hepynet_dir = repo.working_tree_dir
    logger.debug(f"Get hepynet dir: {hepynet_dir}")
    cfg_path = f"{hepynet_dir}/{path}"
    if pathlib.Path(cfg_path).is_file():
        return cfg_path
    elif pathlib.Path(cfg_path).is_dir():
        logger.critical(
            f"Expect a config file path but get directory path: {cfg_path}, please check the config file."
        )
        exit()
    else:
        logger.critical(
            f"No valid config file found at path {cfg_path}, please check the config file."
        )
        exit()


def make_table(data: Iterable, save_dir: ht.pathlike):
    """Makes table for scan meta data and so on.

    Input example:
        data = [
            ["col-1", "col-2", "col-3", "col-4" ],
            [1,2,3,4],
            ["a", "b", "c", "d"]
        ]
    """
    # save csv format
    save_path = save_dir + "/scan_meta_report.csv"
    with open(save_path, "w", newline="") as my_file:
        wr = csv.writer(my_file, quoting=csv.QUOTE_ALL)
        for single_list in data:
            wr.writerow(single_list)



import logging

import matplotlib.pyplot as plt

from hepynet.common.common_utils import get_default_if_none
from hepynet.train import train_utils

logger = logging.getLogger("hepynet")


def get_metrics_band(train_history, metric_name, num_folds):
    # different folds can have different length of metrics
    folds_lengths = list()
    train_folds = list()
    val_folds = list()
    for fold_num in range(num_folds):
        train_fold = train_history[fold_num][metric_name]
        folds_lengths.append(len(train_fold))
        train_folds.append(train_fold)
        val_fold = train_history[fold_num]["val_" + metric_name]
        val_folds.append(val_fold)

    max_len = max(folds_lengths)

    # process train metrics
    (
        train_mean,
        train_low,
        train_high,
    ) = train_utils.merge_unequal_length_arrays(train_folds)
    val_mean, val_low, val_high = train_utils.merge_unequal_length_arrays(
        val_folds
    )

    return {
        "epoch": range(max_len),
        "mean": train_mean,
        "low": train_low,
        "high": train_high,
        "val_mean": val_mean,
        "val_low": val_low,
        "val_high": val_high,
    }


def plot_history(model_wrapper, job_config, save_dir=None):
    """Evaluates training result.

        Args:
            figsize: tuple
                Defines plot size.

        """
    logger.info("Plotting training history")
    plot_config = job_config.apply.cfg_history
    train_history = model_wrapper._train_history
    num_folds = model_wrapper._num_folds
    for metric_key in train_history[0].keys():
        if not metric_key.startswith("val_"):
            plot_metrics(
                metric_key,
                train_history,
                plot_config,
                num_folds=num_folds,
                save_dir=save_dir,
            )


def plot_metrics(
    metric_name,
    train_history,
    plot_config,
    num_folds: int = 1,
    save_dir: str = ".",
) -> None:
    logger.info(f"> Plotting {metric_name}")
    plot_dict = get_metrics_band(train_history, metric_name, num_folds)
    # plot
    fig, ax = plt.subplots()
    ax.plot(plot_dict["epoch"], plot_dict["mean"], label="train")
    if num_folds > 1:
        ax.fill_between(
            plot_dict["epoch"],
            plot_dict["low"],
            plot_dict["high"],
            alpha=0.2,
            label="k-folds train",
        )
    ax.plot(plot_dict["epoch"], plot_dict["val_mean"], label="val")
    if num_folds > 1:
        ax.fill_between(
            plot_dict["epoch"],
            plot_dict["val_low"],
            plot_dict["val_high"],
            alpha=0.2,
            label="k-folds val",
        )
    # config
    plot_title = get_default_if_none(plot_config.plot_title, metric_name)
    ax.set_title(plot_title)
    ax.set_ylabel(metric_name)
    ax.set_xlabel("epoch")
    ax.legend()
    ax.grid()
    save_format = get_default_if_none(plot_config.save_format, "png")
    fig.savefig(f"{save_dir}/history_{metric_name}.{save_format}")



import csv
import logging
import math
import pathlib

import atlas_mpl_style as ampl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.ticker import NullFormatter

import hepynet.common.hepy_type as ht
from hepynet.common import common_utils
from hepynet.data_io import array_utils

logger = logging.getLogger("hepynet")


def calculate_asimov(sig: float, bkg: float):
    return math.sqrt(2 * ((sig + bkg) * math.log(1 + sig / bkg) - sig))


def calculate_significance(
    sig: float,
    bkg: float,
    sig_total: float = None,
    bkg_total: float = None,
    algo="asimov",
):
    """Returns asimov significance"""
    # check input
    if sig <= 0 or bkg <= 0 or sig_total <= 0 or bkg_total <= 0:
        logger.warning(
            "non-positive value found during significance calculation, using default value 0."
        )
        return 0
    if "_rel" in algo:
        if not sig_total:
            logger.error(
                "sig_total or bkg_total value is not specified to calculate relative type significance, please check input."
            )
        if not bkg_total:
            logger.error(
                "sig_total or bkg_total value is not specified to calculate relative type significance, please check input."
            )
    # calculation
    if algo == "asimov":
        return calculate_asimov(sig, bkg)
    elif algo == "s_b":
        return sig / bkg
    elif algo == "s_sqrt_b":
        return sig / math.sqrt(bkg)
    elif algo == "s_sqrt_sb":
        return sig / math.sqrt(sig + bkg)
    elif algo == "asimov_rel":
        return calculate_asimov(sig, bkg) / calculate_asimov(
            sig_total, bkg_total
        )
    elif algo == "s_b_rel":
        return (sig / sig_total) / (bkg / bkg_total)
    elif algo == "s_sqrt_b_rel":
        return (sig / sig_total) / math.sqrt(bkg / bkg_total)
    elif algo == "s_sqrt_sb_rel":
        return (sig / sig_total) / math.sqrt(
            (bkg + sig) / (sig_total + bkg_total)
        )
    else:
        logger.warning(
            "Unrecognized significance algorithm, will use default 'asimov'"
        )
        return calculate_asimov(sig, bkg)


def get_significances(
    df: pd.DataFrame,
    job_config: ht.config,
    significance_algo: str = "asimov",
    multi_class_cut_branch: int = 0,
):
    """Gets significances scan arrays.

    Return:
        Tuple of 4 arrays: (
            threshold array,
            significances,
            sig_total_weight_above_threshold,
            bkg_total_weight_above_threshold,
            )

    """
    ic = job_config.input.clone()
    # prepare signal
    sig_df = array_utils.extract_sig_df(df)
    sig_predictions = sig_df[["y_pred"]].values
    sig_predictions = sig_predictions[:, multi_class_cut_branch]
    # prepare background
    bkg_df = array_utils.extract_bkg_df(df)
    bkg_predictions = bkg_df[["y_pred"]].values
    bkg_predictions = bkg_predictions[:, multi_class_cut_branch]
    # prepare thresholds
    bin_array = np.array(range(-1000, 1000))
    thresholds = 1.0 / (1.0 + 1.0 / np.exp(bin_array * 0.02))
    thresholds = np.insert(thresholds, 0, 0)
    # scan
    significances = []
    plot_thresholds = []
    sig_above_threshold = []
    bkg_above_threshold = []
    sig_weights = sig_df["weight"]
    bkg_weights = bkg_df["weight"]
    total_sig_weight = np.sum(sig_weights)
    total_bkg_weight = np.sum(bkg_weights)
    for dnn_cut in thresholds:
        sig_ids_passed = sig_predictions > dnn_cut
        total_sig_weights_passed = np.sum(sig_weights[sig_ids_passed])
        bkg_ids_passed = bkg_predictions > dnn_cut
        total_bkg_weights_passed = np.sum(bkg_weights[bkg_ids_passed])
        if total_bkg_weights_passed > 0 and total_sig_weights_passed > 0:
            plot_thresholds.append(dnn_cut)
            current_significance = calculate_significance(
                total_sig_weights_passed,
                total_bkg_weights_passed,
                sig_total=total_sig_weight,
                bkg_total=total_bkg_weight,
                algo=significance_algo,
            )
            # current_significance = total_sig_weights_passed / total_bkg_weights_passed
            significances.append(current_significance)
            sig_above_threshold.append(total_sig_weights_passed)
            bkg_above_threshold.append(total_bkg_weights_passed)
    total_sig_weight = np.sum(sig_weights)
    total_bkg_weight = np.sum(bkg_weights)
    return (
        plot_thresholds,
        significances,
        sig_above_threshold,
        bkg_above_threshold,
    )


def plot_significance_scan(
    df: pd.DataFrame, job_config: ht.config, save_dir: ht.pathlike
) -> None:
    """Shows significance change with threshold.

    Note:
        significance is calculated by s/sqrt(b)
    """
    ac = job_config.apply.clone()
    significance_algo = ac.cfg_significance_scan.significance_algo
    logger.info("Plotting significance scan.")
    (
        plot_thresholds,
        significances,
        sig_above_threshold,
        bkg_above_threshold,
    ) = get_significances(df, job_config, significance_algo=significance_algo)

    significances_no_nan = np.nan_to_num(significances)
    max_significance = np.amax(significances_no_nan)
    index = np.argmax(significances_no_nan)
    max_significance_threshold = plot_thresholds[index]
    max_significance_sig_total = sig_above_threshold[index]
    max_significance_bkg_total = bkg_above_threshold[index]
    total_sig_weight = sig_above_threshold[0]
    total_bkg_weight = bkg_above_threshold[0]
    # make plots
    # plot original significance
    original_significance = calculate_significance(
        total_sig_weight,
        total_bkg_weight,
        sig_total=total_sig_weight,
        bkg_total=total_bkg_weight,
        algo=significance_algo,
    )
    fig, ax = plt.subplots()
    ax.set_title("significance scan")
    ax.axhline(y=original_significance, color="grey", linestyle="--")
    # significance scan curve
    ax.plot(
        plot_thresholds,
        significances_no_nan,
        color="r",
        label=significance_algo,
    )
    # signal/background events scan curve
    ax2 = ax.twinx()
    max_sig_events = sig_above_threshold[0]
    max_bkg_events = bkg_above_threshold[0]
    sig_eff_above_threshold = np.array(sig_above_threshold) / max_sig_events
    bkg_eff_above_threshold = np.array(bkg_above_threshold) / max_bkg_events
    ax2.plot(
        plot_thresholds, sig_eff_above_threshold, color="orange", label="sig"
    )
    ax2.plot(
        plot_thresholds, bkg_eff_above_threshold, color="blue", label="bkg"
    )
    ax2.set_ylabel("sig(bkg) ratio after cut")
    # reference threshold
    ax.axvline(x=max_significance_threshold, color="green", linestyle="-.")
    # more infomation
    content = (
        "best threshold:"
        + str(
            common_utils.get_significant_digits(max_significance_threshold, 6)
        )
        + "\nmax significance:"
        + str(common_utils.get_significant_digits(max_significance, 6))
        + "\nbase significance:"
        + str(common_utils.get_significant_digits(original_significance, 6))
        + "\nsig events above threshold:"
        + str(
            common_utils.get_significant_digits(max_significance_sig_total, 6)
        )
        + "\nbkg events above threshold:"
        + str(
            common_utils.get_significant_digits(max_significance_bkg_total, 6)
        )
    )
    ax.text(
        0.05,
        0.05,
        content,
        verticalalignment="bottom",
        horizontalalignment="left",
        transform=ax.transAxes,
        # color="green",
        # fontsize=12,
    )
    # set up plot
    ax.set_title("significance scan")
    ax.set_xscale("logit")
    ax.set_xlabel("DNN score threshold")
    if "rel" in significance_algo:
        ax.set_ylabel("significance ratio")
    else:
        ax.set_ylabel("significance")
    ax.set_ylim(bottom=0)
    ax.locator_params(nbins=10, axis="x")
    ax.yaxis.set_minor_formatter(NullFormatter())
    ax.legend(loc="center left")
    ax2.legend(loc="center right")
    # ax2.set_yscale("log")
    _, y_max = ax.get_ylim()
    ax.set_xlim(1e-3, 1 - 1e-3)
    ax.set_ylim(0, y_max * 1.4)
    _, y_max = ax2.get_ylim()
    ax2.set_ylim(0, y_max * 1.4)
    if ac.plot_atlas_label:
        ampl.plot.draw_atlas_label(
            0.05, 0.95, ax=ax, **(ac.atlas_label.get_config_dict())
        )
    fig_save_path = save_dir / "significance_scan_.png"
    fig.savefig(fig_save_path)

    # collect meta data
    # model_wrapper.original_significance = original_significance
    # model_wrapper.max_significance = max_significance
    # model_wrapper.max_significance_threshold = max_significance_threshold

    # make extra cut table 0.1, 0.2 ... 0.8, 0.9
    # make table for different DNN cut scores
    save_path = save_dir / "scan_DNN_cut.csv"
    with open(save_path, "w", newline="") as file:
        writer = csv.writer(file)
        row_list = [
            [
                "DNN cut",
                "sig events",
                "sig efficiency",
                "bkg events",
                "bkg efficiency",
                "significance",
            ]
        ]
        for index in range(1, 100):
            dnn_cut = (100 - index) / 100.0
            threshold_id = (
                np.abs(np.array(plot_thresholds) - dnn_cut)
            ).argmin()
            sig_events = sig_above_threshold[threshold_id]
            sig_eff = sig_eff_above_threshold[threshold_id]
            bkg_events = bkg_above_threshold[threshold_id]
            bkg_eff = bkg_eff_above_threshold[threshold_id]
            significance = significances[threshold_id]
            new_row = [
                dnn_cut,
                sig_events,
                sig_eff,
                bkg_events,
                bkg_eff,
                significance,
            ]
            row_list.append(new_row)
        row_list.append([""])
        row_list.append(
            [
                "total sig",
                max_sig_events,
                "total bkg",
                max_bkg_events,
                "base significance",
                original_significance,
            ]
        )
        writer.writerows(row_list)
    # make table for different sig efficiency
    save_path = save_dir / "scan_sig_eff.csv"
    with open(save_path, "w", newline="") as file:
        writer = csv.writer(file)
        row_list = [
            [
                "DNN cut",
                "sig events",
                "sig efficiency",
                "bkg events",
                "bkg efficiency",
                "significance",
            ]
        ]
        for index in range(1, 100):
            sig_eff_cut = (100 - index) / 100.0
            threshold_id = (
                np.abs(np.array(sig_eff_above_threshold) - sig_eff_cut)
            ).argmin()
            dnn_cut = plot_thresholds[threshold_id]
            sig_events = sig_above_threshold[threshold_id]
            sig_eff = sig_eff_cut
            bkg_events = bkg_above_threshold[threshold_id]
            bkg_eff = bkg_eff_above_threshold[threshold_id]
            significance = significances[threshold_id]
            new_row = [
                dnn_cut,
                sig_events,
                sig_eff,
                bkg_events,
                bkg_eff,
                significance,
            ]
            row_list.append(new_row)
        row_list.append([""])
        row_list.append(
            [
                "total sig",
                max_sig_events,
                "total bkg",
                max_bkg_events,
                "base significance",
                original_significance,
            ]
        )
        writer.writerows(row_list)
    # make table for different bkg efficiency
    save_path = save_dir / "scan_bkg_eff.csv"
    with open(save_path, "w", newline="") as file:
        writer = csv.writer(file)
        row_list = [
            [
                "DNN cut",
                "sig events",
                "sig efficiency",
                "bkg events",
                "bkg efficiency",
                "significance",
            ]
        ]
        for index in range(1, 100):
            bkg_eff_cut = (100 - index) / 100.0
            threshold_id = (
                np.abs(np.array(bkg_eff_above_threshold) - bkg_eff_cut)
            ).argmin()
            dnn_cut = plot_thresholds[threshold_id]
            sig_events = sig_above_threshold[threshold_id]
            sig_eff = sig_eff_above_threshold[threshold_id]
            bkg_events = bkg_above_threshold[threshold_id]
            bkg_eff = bkg_eff_cut
            significance = significances[threshold_id]
            new_row = [
                dnn_cut,
                sig_events,
                sig_eff,
                bkg_events,
                bkg_eff,
                significance,
            ]
            row_list.append(new_row)
        row_list.append([""])
        row_list.append(
            [
                "total sig",
                max_sig_events,
                "total bkg",
                max_bkg_events,
                "base significance",
                original_significance,
            ]
        )
        writer.writerows(row_list)



import logging
import pathlib
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from hepynet.common import config_utils
from hepynet.data_io import numpy_io
from hepynet.train import hep_model

logger = logging.getLogger("hepynet")


def create_epoch_subdir(save_dir, epoch, n_digit) -> pathlib.Path:
    if save_dir is None:
        logger.error(f"Invalid save_dir: {save_dir}")
        return None
    if epoch is not None:
        sub_dir = pathlib.Path(f"{save_dir}/epoch_{str(epoch).zfill(n_digit)}")
    else:
        sub_dir = pathlib.Path(f"{save_dir}/epoch_final")
    sub_dir.mkdir(parents=True, exist_ok=True)
    return sub_dir


def dump_fit_df(
    model_wrapper: hep_model.Model_Base,
    df_raw,
    df_train,
    job_config,
    save_dir="./",
):
    ic = job_config.input.clone()
    tc = job_config.train.clone()
    ac = job_config.apply.clone()

    sample_list = ic.sig_list + ic.bkg_list
    if ic.apply_data:
        sample_list += ic.data_list

    platform_meta = config_utils.load_current_platform_meta()
    data_path = platform_meta["data_path"]
    if not data_path:
        save_dir = pathlib.Path(save_dir)
    else:
        save_dir = pathlib.Path(data_path) / save_dir
    pathlib.Path(save_dir).mkdir(parents=True, exist_ok=True)
    logger.info(f"> Saving dataframes to: {save_dir}")

    for sample in sample_list:
        logger.info(f"> > processing: {sample}")
        dump_branches = ac.fit_df.branches + ["weight"]
        # prepare contents
        dump_df: pd.DataFrame = df_raw.loc[
            df_raw["sample_name"] == sample, dump_branches
        ]
        input_df = df_train.loc[
            df_train["sample_name"] == sample, ic.selected_features
        ]
        predictions, _, _ = k_folds_predict(
            model_wrapper.get_model(), input_df.values, silence=True
        )
        # dump
        if len(tc.output_bkg_node_names) == 0:
            dump_df["dnn_out_sig"] = predictions
        else:
            for i, out_node in enumerate(["sig"] + tc.output_bkg_node_names):
                out_node = out_node.replace("+", "_")
                branch_name = f"dnn_out_{out_node}"
                dump_df[branch_name] = predictions[:, i]
        dump_df.reset_index(inplace=True)
        save_path = save_dir / f"{sample}.feather"
        dump_df.to_feather(save_path)


def dump_fit_npy(
    model_wrapper: hep_model.Model_Base,
    df_raw,
    df_train,
    job_config,
    npy_dir="./",
):
    ic = job_config.input.clone()
    tc = job_config.train.clone()
    ac = job_config.apply.clone()

    sample_list = ic.sig_list + ic.bkg_list
    if ic.apply_data:
        sample_list += ic.data_list

    platform_meta = config_utils.load_current_platform_meta()
    data_path = platform_meta["data_path"]
    if not data_path:
        save_dir = pathlib.Path(npy_dir)
    else:
        save_dir = pathlib.Path(data_path) / npy_dir
    pathlib.Path(save_dir).mkdir(parents=True, exist_ok=True)
    logger.info(f"> Arrays to be saved to {save_dir}")

    for sample in sample_list:
        dump_branches = ac.cfg_fit_npy.fit_npy_branches + ["weight"]
        # prepare contents
        dump_df = df_raw.loc[df_raw["sample_name"] == sample, dump_branches]
        input_df = df_train.loc[
            df_train["sample_name"] == sample, ic.selected_features
        ]
        predictions, _, _ = k_folds_predict(
            model_wrapper.get_model(), input_df.values
        )
        # dump
        for branch in dump_branches:
            branch_content = dump_df[branch].values
            save_path = f"{save_dir}/{sample}_{branch}.npy"
            numpy_io.save_npy_array(branch_content, save_path)
        if len(tc.output_bkg_node_names) == 0:
            save_path = f"{save_dir}/{sample}_dnn_out.npy"
            numpy_io.save_npy_array(predictions, save_path)
        else:
            for i, out_node in enumerate(["sig"] + tc.output_bkg_node_names):
                out_node = out_node.replace("+", "_")
                save_path = f"{save_dir}/{sample}_dnn_out_{out_node}.npy"
                numpy_io.save_npy_array(predictions[:, i], save_path)


def k_folds_predict(
    k_fold_models, x, silence=False
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    y_pred_k_folds = list()
    if not silence:
        logger.info("Predicting y scores")
    num_folds = len(k_fold_models)
    for fold_num, fold_model in enumerate(k_fold_models):
        if not silence:
            logger.info(
                f"> Predicting with model in fold {fold_num + 1}/{num_folds}"
            )
        y_fold_pred = fold_model.predict(x)
        y_pred_k_folds.append(y_fold_pred)
    y_pred_mean = np.mean(y_pred_k_folds, axis=0)
    y_pred_max = np.maximum.reduce(y_pred_k_folds)
    y_pred_min = np.minimum.reduce(y_pred_k_folds)
    return y_pred_mean, y_pred_min, y_pred_max


def paint_bars(
    ax,
    data,
    labels: list,
    weights,
    bins: int,
    range: tuple,
    title: str = None,
    x_label: str = None,
    y_label: str = None,
    x_unit: str = None,
    x_scale: float = None,
    density: bool = False,
    use_error: bool = False,
    color: str = None,
    fmt: str = ".k",
) -> None:
    """Plot with vertical bar, can be used for data display.

        Note:
        According to ROOT:
        "The error per bin will be computed as sqrt(sum of squares of weight) for each bin."

    """
    plt.ioff()
    # Check input
    data_1dim = np.array([])
    weight_1dim = np.array([])
    if isinstance(data, np.ndarray):
        data = [data]
        weights = [weights]
    for datum, weight in zip(data, weights):
        assert isinstance(
            datum, np.ndarray
        ), "data element should be numpy array."
        assert isinstance(
            weight, np.ndarray
        ), "weights element should be numpy array."
        assert (
            datum.shape == weight.shape
        ), "Input weights should be None or have same type as arrays."
        if len(data_1dim) == 0:
            data_1dim = datum
            weight_1dim = weight
        else:
            data_1dim = np.concatenate((data_1dim, datum))
            weight_1dim = np.concatenate((weight_1dim, weight))

    # Scale x axis
    if x_scale is not None:
        data_1dim = data_1dim * x_scale
    # Make bar plot
    # get bin error and edges
    plot_ys, _ = np.histogram(
        data_1dim, bins=bins, range=range, weights=weight_1dim, density=density
    )
    sum_weight_squares, bin_edges = np.histogram(
        data_1dim, bins=bins, range=range, weights=np.power(weight_1dim, 2)
    )
    if density:
        error_scale = 1 / (np.sum(weight_1dim) * (range[1] - range[0]) / bins)
        errors = np.sqrt(sum_weight_squares) * error_scale
    else:
        errors = np.sqrt(sum_weight_squares)
    # Only plot ratio when bin is not 0.
    bin_centers = np.array([])
    bin_ys = np.array([])
    bin_yerrs = np.array([])
    for i, y1 in enumerate(plot_ys):
        if y1 != 0:
            ele_center = np.array([0.5 * (bin_edges[i] + bin_edges[i + 1])])
            bin_centers = np.concatenate((bin_centers, ele_center))
            ele_y = np.array([y1])
            bin_ys = np.concatenate((bin_ys, ele_y))
            ele_yerr = np.array([errors[i]])
            bin_yerrs = np.concatenate((bin_yerrs, ele_yerr))
    # plot bar
    bin_size = bin_edges[1] - bin_edges[0]
    if use_error:
        ax.errorbar(
            bin_centers,
            bin_ys,
            xerr=bin_size / 2.0,
            yerr=bin_yerrs,
            fmt=fmt,
            label=labels,
            color=color,
            markerfacecolor=color,
            markeredgecolor=color,
        )
    else:
        ax.errorbar(
            bin_centers,
            bin_ys,
            xerr=bin_size / 2.0,
            yerr=None,
            fmt=fmt,
            label=labels,
            color=color,
            markerfacecolor=color,
            markeredgecolor=color,
        )
    # Config
    if title is not None:
        ax.set_title(title)
    if x_label is not None:
        if x_unit is not None:
            ax.set_xlabel(x_label + "/" + x_unit)
        else:
            ax.set_xlabel(x_label)
    else:
        if x_unit is not None:
            ax.set_xlabel(x_unit)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if range is not None:
        ax.axis(xmin=range[0], xmax=range[1])
    ax.legend(loc="upper right")



import glob
import logging
import platform
import re
import socket
from typing import Any

logger = logging.getLogger("hepynet")


def get_current_platform_name() -> str:
    """Returns the name of the current platform.

    Returns:
        str: name of current platform
    """
    return platform.platform()


def get_current_hostname() -> str:
    """Returns the hostname of current machine

    Returns:
        str: current hostname
    """
    return socket.gethostname()


def get_default_if_none(input_var: Any, default_value: Any):
    if input_var is None:
        return default_value
    else:
        return input_var


def get_newest_file_version(
    path_pattern: str,
    n_digit: int = 2,
    ver_num: int = None,
    use_existing: bool = False,
):
    """Check existed file and return last available file path with version.

    Version range 00 -> 99 (or 999)
    If reach limit, last available version will be used. 99 (or 999)

    """
    # Return file path if ver_num is given
    if ver_num is not None:
        return {
            "ver_num": ver_num,
            "path": path_pattern.format(str(ver_num).zfill(n_digit)),
        }
    # Otherwise try to find ver_num
    path_list = glob.glob(path_pattern.format("*"))
    path_list = sorted(path_list)
    if len(path_list) < 1:
        if use_existing:
            logger.debug(
                f"Can't find existing file with path pattern: {path_pattern}, returning empty."
            )
            return {}
        else:
            ver_num = 0
            path = path_pattern.format(str(0).zfill(n_digit))
    else:
        path = path_list[-1]  # Choose the last match
        version_tag_search = re.compile("v(" + "\d" * n_digit + ")")
        ver_num = int(version_tag_search.search(path).group(1))
        if not use_existing:
            ver_num += 1
            path = path_pattern.format(str(ver_num).zfill(n_digit))

    return {
        "ver_num": ver_num,
        "path": path,
    }


def get_significant_digits(number: float, n_digits: int):
    if round(number) == number:
        m = len(str(number)) - 1 - n_digits
        if number / (10 ** m) == 0.0:
            return number
        else:
            return float(int(number) / (10 ** m) * (10 ** m))
    if len(str(number)) > n_digits + 1:
        return round(number, n_digits - len(str(int(number))))
    else:
        return number



from __future__ import nested_scopes

import copy
import logging
import pathlib
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import yaml
from sklearn.model_selection import StratifiedShuffleSplit

from hepynet.common import config_utils
from hepynet.data_io import numpy_io

logger = logging.getLogger("hepynet")


class Feedbox(object):
    """DNN inputs management class."""

    def __init__(
        self, job_config: config_utils.Hepy_Config, norm_dict=None,
    ):
        # get config
        self._job_config = job_config.clone()
        self._data_dir = pathlib.Path(numpy_io.get_data_dir())
        jc = self._job_config.job.clone()
        self._array_prepared = False

        # get norm dict
        if norm_dict is None:
            if jc.job_name == "apply":
                logger.warning(
                    f"No norm_dict found, but this is not expected as the job type is 'apply'! The normalization parameters may not be consistent with the training!"
                )
            self._norm_dict = {}
        else:
            self._norm_dict = copy.deepcopy(norm_dict)

        self._array_prepared = True

    def get_raw_df(self) -> pd.DataFrame:
        logger.info("Loading raw input DataFrame...")
        ic = self._job_config.input.clone()
        raw_df: pd.DataFrame = pd.read_feather(self._data_dir / ic.input_path)
        # select events in sig/bkg sample_list
        sample_list = ic.sig_list + ic.bkg_list + ic.data_list
        logger.info(f"> Loading samples in list: {sample_list}")
        # raw_df = raw_df[raw_df["sample_name"].isin(sample_list)]
        # raw_df.drop(raw_df[~raw_df["sample_name"].isin(sample_list)], inplace=True)
        raw_df.drop(
            raw_df.index[~raw_df["sample_name"].isin(sample_list)],
            inplace=True,
        )
        # select events by extra cut features
        if ic.cut_expression:
            logger.info(
                f"> Cutting inputs according to expression: {ic.cut_expression}"
            )
            raw_df.query(ic.cut_expression, inplace=True)
        # return
        raw_df.reset_index(drop=True, inplace=True)
        logger.info("> Successfully loaded raw input DataFrame...")
        return raw_df

    def get_processed_df(
        self, raw_df: pd.DataFrame = None, keep_unreset: bool = False
    ):
        logger.info("Loading processed input DataFrame...")
        ic = self._job_config.input.clone()
        if raw_df is None:
            out_df: pd.DataFrame = self.get_raw_df()
        else:
            out_df: pd.DataFrame = raw_df.copy()
        feature_include = ic.selected_features + ic.validation_features
        # overwrite physic parameter
        if ic.reset_feature_overwrite:
            physic_para = ic.reset_feature_name
            physic_para_value = ic.reset_feature_overwrite_value
            logger.info(f"> Setting physic para to {physic_para_value}")
            if not isinstance(physic_para_value, (int, float)):
                logger.error(
                    f"Invalid physic_para_value type: {type(physic_para_value)}"
                )
                exit()
            out_df.loc[
                out_df["is_sig"] == False, physic_para
            ] = physic_para_value
        # reshape inputs
        if ic.reshape_input:
            logger.info("> Reshaping inputs")
            # check missing norm parameters
            missing_norms = []
            norm_alias = ic.feature_norm_alias.get_config_dict()
            for feature in feature_include:
                if feature not in self._norm_dict:
                    found_norm_alias = False
                    # check feature normalization alias
                    if feature in norm_alias:
                        feature_alias = norm_alias[feature]
                        if feature_alias in self._norm_dict:
                            logger.debug(f"Found norm_alias for: {feature}")
                            found_norm_alias = True
                            self._norm_dict[feature] = copy.deepcopy(
                                self._norm_dict[feature_alias]
                            )
                        else:
                            logger.warn(
                                f"Specified but can't find norm_alias for: {feature}"
                            )
                    if not found_norm_alias:
                        missing_norms.append(feature)
            # recalculate missing norm parameters
            if len(missing_norms) > 0:
                logger.debug(
                    f"Recalculating normalization parameters for {missing_norms}"
                )
                self.update_norm_dict(missing_norms)
            # reshape inputs
            for feature in feature_include:
                if feature in self._norm_dict:
                    f_mean = self._norm_dict[feature]["mean"]
                    f_var = self._norm_dict[feature]["variance"]
                    new_values = (
                        out_df[feature].to_numpy("float32") - f_mean
                    ) / np.sqrt(f_var)
                    out_df.loc[:, feature].update(new_values)
                else:
                    logger.debug(
                        f"{feature} is not in self._norm_dict, ignoring reshape process."
                    )
        # deal with negative weights
        nwt_method = ic.negative_weight_process
        logger.info(f"> Processing negative weights with method: {nwt_method}")
        if nwt_method == "to_zero":
            out_df.loc[out_df["weight"] < 0, "weight"] = 0
        elif nwt_method == "to_positive":
            out_df.loc[out_df["weight"] < 0, "weight"].update(
                out_df.loc[out_df["weight"] < 0, "weight"].abs()
            )
        elif nwt_method == "keep":
            pass
        else:
            logger.warn(f"> Unknown ic.negative_weight_process {nwt_method}")
        # reweight inputs
        if ic.reweight_input:
            logger.info("> Reweighting inputs")
            # reweight signal
            self.reweight_df(
                out_df, ic.sig_list, ic.sig_key, ic.sig_sumofweight, True, True
            )
            # reweight background
            self.reweight_df(
                out_df,
                ic.bkg_list,
                ic.bkg_key,
                ic.bkg_sumofweight,
                True,
                False,
            )
        # reset physic para for pDNN
        physic_para = ic.reset_feature_name
        if keep_unreset:
            unreset_physic_para_values = out_df[physic_para].values
            out_df[physic_para + "_unreset"] = unreset_physic_para_values
        if ic.reset_feature:
            logger.info(
                f"> Resetting physic para {physic_para} distribution for pDNN"
            )
            ref_df = out_df.loc[
                out_df["is_sig"] == True, [physic_para, "weight"]
            ]
            ref_array = ref_df[physic_para]
            ref_weight = ref_df["weight"]
            ref_weight_positive = ref_weight.copy()
            ref_weight_positive[ref_weight_positive < 0] = 0
            sump = ref_weight_positive.sum()
            reset_size = len(out_df.loc[out_df["is_sig"] == False].index)
            reset_values = np.random.choice(
                ref_array.to_numpy("float32"),
                size=reset_size,
                p=(1 / sump) * ref_weight_positive.to_numpy("float32"),
            )
            out_df.loc[out_df["is_sig"] == False, physic_para] = reset_values

        # set y
        logger.info("> Setting up y values")
        out_df.loc[:, "y"] = 0
        out_df.loc[out_df["is_sig"] == True, "y"] = 1
        # tag train / test
        logger.info("> Tagging train / test")
        sss = StratifiedShuffleSplit(n_splits=1, test_size=ic.test_rate)
        y_arr = out_df["y"].to_numpy("float32")
        for train_index, _ in sss.split(y_arr, y_arr):
            break
        out_df.loc[:, "is_train"] = False
        out_df.loc[train_index, "is_train"] = True

        logger.info("> Successfully loaded processed input DataFrame...")
        return out_df

    def get_job_config(self):
        return self._job_config

    def get_norm_dict(self):
        return self._norm_dict

    def reweight_df(
        self,
        df: pd.DataFrame,
        sample_list,
        array_key,
        sumofweight,
        is_mc,
        is_sig,
    ):
        total_wt = df.loc[
            (df["is_mc"] == is_mc)
            & (df["is_sig"] == is_sig)
            & df["sample_name"].isin(sample_list),
            "weight",
        ].sum()
        if array_key in sample_list:
            sample_list = [array_key]
        for sample in sample_list:
            sample_wt = df.loc[df["sample_name"] == sample, "weight"]
            sample_sumw = sample_wt.sum()
            if sample_sumw <= 0:
                continue
            # get norm factor
            norm_factor = 1
            if array_key == "all":
                norm_factor = sumofweight / total_wt
            elif array_key == "all_norm":
                n_samples = len(sample_list)
                norm_factor = (sumofweight / n_samples) / sample_sumw
            elif array_key == sample:
                norm_factor = sumofweight / sample_sumw
            else:
                logger.warn(
                    f"Unknown array_key {array_key} for sample {sample}"
                )
            df.loc[df["sample_name"] == sample, "weight"] = (
                sample_wt * norm_factor
            )

    def update_norm_dict(self, features: Optional[List[str]] = None):
        ic = self._job_config.input.clone()
        feature_list = list()
        if features is not None:
            feature_list = features
        else:
            feature_list = ic.selected_features + ic.validation_features
        feature_list = list(set().union(feature_list))  # remove duplicates
        df = self.get_raw_df()
        weight_array = df.loc[
            (df["is_mc"] == True) & (df["is_sig"] == False), "weight"
        ].to_numpy("float32")
        for feature in feature_list:
            feature_array = df.loc[
                (df["is_mc"] == True) & (df["is_sig"] == False), feature
            ].to_numpy("float32")
            mean = np.average(feature_array, weights=weight_array)
            variance = np.average(
                (feature_array - mean) ** 2, weights=weight_array
            )
            self._norm_dict[feature] = {"mean": mean, "variance": variance}
            logger.info(
                f"> > Recalculated norm factors for feature {feature} mean: {mean}, variance: {variance}"
            )



# -*- coding: utf-8 -*-
import logging
import math
import os
import pathlib
from typing import List, Optional

import numpy as np
import ray
import yaml
from ray import tune
from ray.tune import schedulers, stopper
from sklearn.model_selection import StratifiedKFold

import hepynet.common.hepy_type as ht
from hepynet.train import hep_model

logger = logging.getLogger("hepynet")


def get_model_class(model_class: str):
    if model_class == "Model_Sequential_Flat":
        return hep_model.Model_Sequential_Flat
    else:
        logger.critical(f"Unknown model class: {model_class}")
        exit(1)


def get_mean_var(
    array: np.ndarray,
    axis: Optional[int] = None,
    weights: Optional[np.ndarray] = None,
):
    """Calculate average and variance of an array."""
    average = np.average(array, axis=axis, weights=weights)
    variance = np.average((array - average) ** 2, axis=axis, weights=weights)
    if 0 in variance:
        logger.warn("Encountered 0 variance, adding shift value 0.000001")
    return average, variance + 0.000001


def get_single_hyper(hyper_config: ht.sub_config):
    """Determines one hyperparameter
    
    Note:
        If the hyper-parameter is a tuning dimension (with spacer specified), 
        spacer from ray.tune will be used to set up the dimension.
        If the hyper-parameter is a normal value, it will be returned directly.

    """
    if hasattr(hyper_config, "spacer"):
        paras = hyper_config.paras.get_config_dict()
        return getattr(tune, hyper_config.spacer)(**paras)
    else:
        return hyper_config


def get_train_val_indices(
    x: np.ndarray,
    y: np.ndarray,
    wt: np.ndarray,
    val_split: float,
    k_folds: Optional[int] = None,
):
    """Gets indices to separates train datasets to train/validation"""
    train_indices_list = list()
    validation_indices_list = list()
    if isinstance(k_folds, int) and k_folds >= 2:
        # skf = KFold(n_splits=k_folds, shuffle=True)
        skf = StratifiedKFold(n_splits=k_folds, shuffle=True)
        for train_index, val_index in skf.split(x, y):
            np.random.shuffle(train_index)
            np.random.shuffle(val_index)
            train_indices_list.append(train_index)
            validation_indices_list.append(val_index)
    else:
        if isinstance(k_folds, int) and k_folds <= 2:
            logger.error(
                f"Invalid train.k_folds value {k_folds} detected, will not use k-fold validation"
            )
        array_len = len(wt)
        val_index = np.random.choice(
            range(array_len), int(array_len * 1.0 * val_split), replace=False
        )
        train_index = np.setdiff1d(np.array(range(array_len)), val_index)
        np.random.shuffle(train_index)
        np.random.shuffle(val_index)
        train_indices_list.append(train_index)
        validation_indices_list.append(val_index)
    return train_indices_list, validation_indices_list


def merge_unequal_length_arrays(array_list: List[np.ndarray]):
    """Merges arrays with unequal length to average/min/max

    Note:
        mainly used to deal with k-fold results with early-stopping (which
        results in unequal length of results as different folds may stop at
        different epochs) enabled

    """
    folds_lengths = list()
    for single_array in array_list:
        folds_lengths.append(len(single_array))
    max_len = max(folds_lengths)

    mean = list()
    low = list()
    high = list()
    for i in range(max_len):
        sum_value = 0
        min_value = math.inf
        max_value = -math.inf
        num_values = 0
        for fold_num, single_array in enumerate(array_list):
            if i < folds_lengths[fold_num]:
                ele = single_array[i]
                sum_value += ele
                num_values += 1
                if ele < min_value:
                    min_value = ele
                if ele > max_value:
                    max_value = ele
        mean_value = sum_value / num_values
        mean.append(mean_value)
        low.append(min_value)
        high.append(max_value)
    return mean, low, high


def ray_tune(model_wrapper, job_config: ht.config, resume: bool = False):
    """Performs automatic hyper-parameters tuning with Ray"""
    # initialize
    tuner = job_config.tune.clone().tuner
    log_dir = pathlib.Path(job_config.run.save_sub_dir) / "tmp_log"
    log_dir.mkdir(parents=True, exist_ok=True)
    if os.name == "posix":
        logger.info(f"Ignoring tune.tmp.tmp_dir setting on Unix OS")
        ray.init(**(tuner.init.get_config_dict()))
    else:
        ray.init(
            _temp_dir=str(job_config.tune.tmp_dir),
            **(tuner.init.get_config_dict()),
        )
    # set up scheduler
    sched_class = getattr(schedulers, tuner.scheduler_class)
    logger.info(f"Setting up scheduler: {tuner.scheduler_class}")
    sched_config = tuner.scheduler.get_config_dict()
    sched = sched_class(**sched_config)
    # set up algorithm
    algo_class = tuner.algo_class
    logger.info(f"Setting up search algorithm: {tuner.algo_class}")
    algo_config = tuner.algo.get_config_dict()
    algo = None
    if algo_class is None:
        algo = None
    elif algo_class == "AxSearch":
        from ray.tune.suggest.ax import AxSearch

        algo = AxSearch(**algo_config)
    elif algo_class == "HyperOptSearch":
        from ray.tune.suggest.hyperopt import HyperOptSearch

        algo = HyperOptSearch(**algo_config)
    elif algo_class == "HEBOSearch":
        from ray.tune.suggest.hebo import HEBOSearch

        algo = HEBOSearch(**algo_config)
    else:
        logger.error(f"Unsupported search algorithm: {algo_class}")
        logger.info(f"Using default value None for search algorithm")
    # set stopper
    if tuner.stopper_class is None:
        stop = None
    else:
        stop = getattr(stopper, tuner.stopper_class)(**tuner.stopper)
    # run
    run_config = (
        tuner.run.get_config_dict()
    )  # important: convert Hepy_Config class to dict
    tune_func = getattr(hep_model, model_wrapper._tune_fun_name)
    analysis = tune.run(
        tune_func,
        name="ray_tunes",
        stop=stop,
        search_alg=algo,
        scheduler=sched,
        config=model_wrapper.get_hypers_tune(),
        local_dir=job_config.run.save_sub_dir,
        resume=resume,
        **run_config,
    )
    logger.info("Best hyperparameters found were:")
    print(yaml.dump(analysis.best_config))

    return analysis



# -*- coding: utf-8 -*-

# TODO: evaluate.py is disables and should be remove once equivalent functions developed

"""Functions for making plots"""

import logging
import math

import numpy as np

from hepynet.data_io import array_utils, feed_box, numpy_io
from hepynet.train import train_utils

try:
    import ROOT

    root_available = True
    from easy_atlas_plot.plot_utils import plot_utils_root, th1_tools
except ImportError:
    root_available = False

logger = logging.getLogger("hepynet")


def plot_2d_density(
    job_wrapper, save_plot=False, save_dir=None, save_file_name="2d_density",
):
    """Plots 2D hist to see event distribution of signal and backgrounds events.

    x-axis will be dnn scores, y-axis will be mass parameter, z-axis is total
    weight shown by color

    """
    ic = job_wrapper.job_config.input
    model_wrapper = job_wrapper.model_wrapper
    feedbox = model_wrapper.feedbox
    model_meta = model_wrapper.model_meta
    # plot signal
    sig_key = model_meta["sig_key"]
    sig_arr_raw, _ = feedbox.get_raw("xs", array_key=sig_key)
    sig_arr, sig_weights = feedbox.get_reshape("xs", array_key=sig_key)
    predict_arr = model_wrapper.get_model().predict(sig_arr)
    if predict_arr.ndim == 2:
        predict_arr = predict_arr[:, 0]
    mass_index = ic.selected_features.index(ic.reset_feature_name)
    x = predict_arr
    y = sig_arr_raw[:, mass_index]
    ## make plot
    plot_canvas = ROOT.TCanvas("2d_density_sig", "2d_density_sig", 800, 600)
    hist_sig = th1_tools.TH2FTool(
        "2d_density_sig",
        "2d_density_sig",
        nbinx=50,
        xlow=0,
        xup=1.0,
        nbiny=50,
        ylow=min(y),
        yup=max(y),
    )
    hist_sig.fill_hist(
        fill_array_x=x, fill_array_y=y, weight_array=sig_weights
    )
    hist_sig.set_canvas(plot_canvas)
    hist_sig.set_palette("kBird")
    hist_sig.update_config("hist", "SetStats", 0)
    hist_sig.update_config("x_axis", "SetTitle", "dnn score")
    hist_sig.update_config("y_axis", "SetTitle", "mass")
    hist_sig.draw("colz")
    hist_sig.save(save_dir=save_dir, save_file_name=save_file_name + "_sig")
    # plot background
    bkg_key = model_meta["bkg_key"]
    bkg_arr_raw, _ = feedbox.get_raw("xb", array_key=bkg_key)
    bkg_arr, bkg_weights = feedbox.get_reshape("xb", array_key=bkg_key)
    predict_arr = model_wrapper.get_model().predict(bkg_arr)
    if predict_arr.ndim == 2:
        predict_arr = predict_arr[:, 0]
    mass_index = ic.selected_features.index(ic.reset_feature_name)
    x = predict_arr
    y = bkg_arr_raw[:, mass_index]
    ## make plot
    plot_canvas = ROOT.TCanvas("2d_density_bkg", "2d_density_bkg", 800, 600)
    hist_bkg = th1_tools.TH2FTool(
        "2d_density_bkg",
        "2d_density_bkg",
        nbinx=50,
        xlow=0,
        xup=1.0,
        nbiny=50,
        ylow=min(y),
        yup=max(y),
    )
    hist_bkg.fill_hist(
        fill_array_x=x, fill_array_y=y, weight_array=bkg_weights
    )
    hist_bkg.set_canvas(plot_canvas)
    hist_bkg.set_palette("kBird")
    hist_bkg.update_config("hist", "SetStats", 0)
    hist_bkg.update_config("x_axis", "SetTitle", "dnn score")
    hist_bkg.update_config("y_axis", "SetTitle", "mass")
    hist_bkg.draw("colz")
    hist_bkg.save(save_dir=save_dir, save_file_name=save_file_name + "_bkg")


def plot_2d_significance_scan(
    job_wrapper,
    save_plot=False,
    save_dir=None,
    save_file_name="2d_significance",
    cut_ranges_dn=None,
    cut_ranges_up=None,
    dnn_cut_min=None,
    dnn_cut_max=None,
    dnn_cut_step=None,
):
    """Makes 2d map of significance"""
    if not (dnn_cut_min and dnn_cut_max and dnn_cut_step):
        logging.warn(
            "No complete dnn_cut specified, using default (min=0.5, max=1, step=0.05)"
        )
        dnn_cut_min = 0.5
        dnn_cut_max = 1
        dnn_cut_step = 0.05
    dnn_cut_list = np.arange(dnn_cut_min, dnn_cut_max, dnn_cut_step)
    w_inputs = []
    logger.info("Making 2d significance scan.")
    ic = job_wrapper.job_config.input
    tc = job_wrapper.job_config.train
    ac = job_wrapper.job_config.apply
    rc = job_wrapper.job_config.run
    bkg_dict = numpy_io.load_npy_arrays(
        rc.npy_path,
        ic.campaign,
        ic.region,
        ic.channel,
        ic.bkg_list,
        ic.selected_features,
        validation_features=ic.validation_features,
        cut_features=ic.cut_features,
        cut_values=ic.cut_values,
        cut_types=ic.cut_types,
    )
    sig_dict = numpy_io.load_npy_arrays(
        rc.npy_path,
        ic.campaign,
        ic.region,
        ic.channel,
        ic.sig_list,
        ic.selected_features,
        validation_features=ic.validation_features,
        cut_features=ic.cut_features,
        cut_values=ic.cut_values,
        cut_types=ic.cut_types,
    )
    for sig_id, scan_sig_key in enumerate(ic.sig_list):
        xs = sig_dict[scan_sig_key]
        m_cut_name = ic.reset_feature_name
        if cut_ranges_dn is None or len(cut_ranges_dn) == 0:
            means, variances = train_utils.get_mean_var(
                xs[m_cut_name], axis=0, weights=xs["weight"]
            )
            m_cut_dn = means[0] - math.sqrt(variances[0])
            m_cut_up = means[0] + math.sqrt(variances[0])
        else:
            m_cut_dn = cut_ranges_dn[sig_id]
            m_cut_up = cut_ranges_up[sig_id]
        feedbox = feed_box.Feedbox(
            sig_dict,
            bkg_dict,
            selected_features=ic.selected_features,
            apply_data=False,
            reshape_array=ic.norm_array,
            reset_mass=ic.reset_feature,
            reset_mass_name=ic.reset_feature_name,
            remove_negative_weight=ic.rm_negative_weight_events,
            cut_features=[m_cut_name, m_cut_name],
            cut_values=[m_cut_dn, m_cut_up],
            cut_types=[">", "<"],
            sig_weight=ic.sig_sumofweight,
            bkg_weight=ic.bkg_sumofweight,
            data_weight=ic.data_sumofweight,
            test_rate=ic.test_rate,
            model_meta=job_wrapper.model_wrapper.model_meta,
            verbose=tc.verbose,
        )
        job_wrapper.model_wrapper.set_inputs(feedbox, apply_data=ac.apply_data)
        significance_algo = ac.cfg_2d_significance_scan["significance_algo"]
        (plot_thresholds, significances, _, _,) = get_significances(
            job_wrapper.model_wrapper,
            ic.sig_key,
            ic.bkg_key,
            significance_algo=significance_algo,
        )
        plot_significances = []
        for dnn_cut in dnn_cut_list:
            threshold_id = (
                np.abs(np.array(plot_thresholds) - dnn_cut)
            ).argmin()
            plot_significances.append(significances[threshold_id])
        w_inputs.append(plot_significances)
    x = []
    y = []
    w = []
    for index, w_input in enumerate(w_inputs):
        new_x = (dnn_cut_list + dnn_cut_step / 2).tolist()
        new_y = [ic.sig_list[index]] * len(w_input)
        new_w = w_input
        logging.debug(f"new x: {new_x}")
        logging.debug(f"new y: {new_y}")
        logging.debug(f"new w: {new_w}")
        if len(x) == 0:
            x = new_x
            y = new_y
            w = new_w
        else:
            x += new_x
            y += new_y
            w += new_w
    # make plot
    plot_canvas = ROOT.TCanvas(
        "2d_significance_c", "2d_significance_c", 800, 600
    )
    hist_sig = th1_tools.TH2FTool(
        "2d_significance",
        "2d_significance",
        nbinx=len(dnn_cut_list),
        xlow=dnn_cut_min,
        xup=dnn_cut_max,
        nbiny=len(ic.sig_list),
        ylow=0,
        yup=len(ic.sig_list),
    )
    hist_sig_text = th1_tools.TH2FTool(
        "2d_significance_text",
        "2d_significance_text",
        nbinx=len(dnn_cut_list),
        xlow=dnn_cut_min,
        xup=dnn_cut_max,
        nbiny=len(ic.sig_list),
        ylow=0,
        yup=len(ic.sig_list),
    )
    hist_sig.fill_hist(fill_array_x=x, fill_array_y=y, weight_array=w)
    hist_sig.set_canvas(plot_canvas)
    hist_sig.set_palette("kBird")
    hist_sig.update_config("hist", "SetStats", 0)
    hist_sig.update_config("x_axis", "SetTitle", "dnn_cut")
    hist_sig.update_config("y_axis", "SetTitle", "mass point")
    hist_sig.update_config("y_axis", "SetLabelOffset", 0.003)
    hist_sig.draw("colz")
    hist_sig_text.fill_hist(
        fill_array_x=x,
        fill_array_y=y,
        weight_array=np.array(w).round(decimals=2),
    )
    hist_sig_text.set_canvas(plot_canvas)
    hist_sig_text.update_config("hist", "SetMarkerSize", 1.8)
    hist_sig_text.draw("text same")
    if save_plot:
        plot_canvas.SaveAs(save_dir + "/" + save_file_name + ".png")



# defines type hints in hepynet
import os
from typing import List, Tuple, Union

import matplotlib.axes._axes as axes

from hepynet.common import config_utils

# from hepynet.train import hep_model  # TODO: this may cause circular import issue

# matplotlib objects
ax = axes.Axes

# python types combination
bound = Union[Tuple[float, float], List[float]]
pathlike = Union[str, os.PathLike]

# hepynet objects
config = config_utils.Hepy_Config
sub_config = config_utils.Hepy_Config_Section
# model = hep_model.Model_Base



import datetime
import logging
import pathlib
import random as python_random
import shutil
import time

import atlas_mpl_style as ampl
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import yaml
from cycler import cycler

import hepynet.common.hepy_type as ht
from hepynet.common import common_utils, config_utils
from hepynet.data_io import feed_box
from hepynet.evaluate import (
    evaluate_utils,
    importance,
    kinematics,
    metrics,
    mva_scores,
    significance,
    train_history,
)
from hepynet.main import job_utils
from hepynet.train import train_utils

logger = logging.getLogger("hepynet")


class job_executor(object):
    """Core class to execute a pdnn job based on given cfg file."""

    def __init__(self, yaml_config_path: ht.pathlike):
        """Initialize executor."""
        self.job_config = None
        self.get_config(yaml_config_path)
        # Set up global plot style and color cycle
        ampl.use_atlas_style(usetex=False)
        color_cycle = self.job_config.apply.color_cycle
        default_cycler = cycler(color=color_cycle)
        plt.rc("axes", prop_cycle=default_cycler)

    def execute_jobs(self, resume: bool = False):
        """Execute all planned jobs.

        TODO: This function is no longer needed as tune job has been designed to
        be an independent job type. Consider merge with execute_single_job in
        future.

        """
        self.job_config.print()
        self.set_save_dir(resume=resume)
        # Execute single job if parameter scan is not needed
        self.execute_single_job(resume=resume)

    def execute_single_job(self, resume: bool = False):
        """Execute single DNN training with given configuration."""
        # Prepare
        jc = self.job_config.job
        rc = self.job_config.run
        if jc.job_type == "apply":
            if rc.load_dir == None:
                rc.load_dir = jc.save_dir

        if jc.fix_rdm_seed:
            self.fix_random_seed()

        # Check best tuned config overwrite
        if self.job_config.config.best_tune_overwrite:
            logger.info("Overwriting training config with best tuned results")
            best_hypers_path = (
                pathlib.Path(rc.save_sub_dir) / "best_hypers.yaml"
            )
            if best_hypers_path.is_file():
                with open(best_hypers_path, "r") as best_hypers_file:
                    best_hypers = yaml.load(
                        best_hypers_file, Loader=yaml.FullLoader
                    )
                    self.job_config.train.update(best_hypers)

        self.set_model()
        self.set_model_input()

        if jc.job_type == "train":
            self.execute_train_job()
        elif jc.job_type == "tune":
            self.execute_tune_job(resume=resume)
        elif jc.job_type == "apply":
            self.execute_apply_job()
        else:
            logger.critical(
                f"job.job_type must be train or apply, {jc.job_type} is not supported"
            )

        # Post procedure
        plt.close("all")

    def execute_train_job(self):
        # Save the final job_config
        rc = self.job_config.run
        yaml_path = pathlib.Path(rc.save_sub_dir) / "train_config.yaml"
        with open(yaml_path, "w") as yaml_file:
            yaml.dump(self.job_config.get_config_dict(), yaml_file, indent=4)
        # Train
        self.model_wrapper.build()
        self.model_wrapper.train()

    def execute_tune_job(self, resume: bool = False):
        # Save the final job_config
        ic = self.job_config.input
        uc = self.job_config.tune
        rc = self.job_config.run
        yaml_path = pathlib.Path(rc.save_sub_dir) / "tune_config.yaml"
        with open(yaml_path, "w") as yaml_file:
            yaml.dump(self.job_config.get_config_dict(), yaml_file, indent=4)

        # Prepare tmp inputs
        feedbox = feed_box.Feedbox(self.job_config.clone())
        ## get input
        input_df = feedbox.get_processed_df(keep_unreset=True)
        cols = ic.selected_features
        ## load and save train/test
        train_index = (
            input_df["is_train"] == True
        )  # index for train and validation
        x = input_df.loc[train_index, cols].values
        y = input_df.loc[train_index, ["y"]].values
        wt = input_df.loc[train_index, "weight"].values
        val_ids = np.random.choice(
            range(len(wt)),
            int(len(wt) * 1.0 * uc.model.val_split),
            replace=False,
        )
        train_ids = np.setdiff1d(np.array(range(len(wt))), val_ids)
        x_train = x[train_ids]
        y_train = y[train_ids]
        wt_train = wt[train_ids]
        x_val = x[val_ids]
        y_val = y[val_ids]
        wt_val = wt[val_ids]
        ## prepare inputs without resetting physic parameter
        physic_para = ic.reset_feature_name
        cols_unreset = list()
        for col in cols:
            if col == physic_para:
                cols_unreset.append(physic_para + "_unreset")
            else:
                cols_unreset.append(col)
        x_unreset = input_df.loc[train_index, cols_unreset].values
        x_train_unreset = x_unreset[train_ids]
        x_val_unreset = x_unreset[val_ids]
        ## remove negative weight events
        if ic.rm_negative_weight_events == True:
            wt_train = wt_train.clip(min=0)
        tune_input_dir = pathlib.Path(rc.tune_input_cache)
        np.save(tune_input_dir / "x_train.npy", x_train)
        np.save(tune_input_dir / "x_train_unreset.npy", x_train_unreset)
        np.save(tune_input_dir / "y_train.npy", y_train)
        np.save(tune_input_dir / "wt_train.npy", wt_train)
        np.save(tune_input_dir / "x_val.npy", x_val)
        np.save(tune_input_dir / "x_val_unreset.npy", x_val_unreset)
        np.save(tune_input_dir / "y_val.npy", y_val)
        np.save(tune_input_dir / "wt_val.npy", wt_val)
        logger.info(f"Temporary tuning input files saved to: {tune_input_dir}")

        # Tuning hypers
        analysis = train_utils.ray_tune(
            self.model_wrapper, self.job_config, resume=resume
        )

        # Save results
        best_config = analysis.best_config
        save_path = pathlib.Path(rc.save_sub_dir) / "best_hypers.yaml"
        with open(save_path, "w") as best_hypers_file:
            yaml.dump(best_config, best_hypers_file, indent=4)
        results = analysis.results
        save_path = pathlib.Path(rc.save_sub_dir) / "tune_results.yaml"
        with open(save_path, "w") as tune_results_file:
            yaml.dump(results, tune_results_file, indent=4)
        save_path = pathlib.Path(rc.save_sub_dir) / "tune_best_trial.yaml"
        with open(save_path, "w") as best_trial_file:
            yaml.dump(analysis.best_trial, best_trial_file, indent=4)

        # Remove temporary tuning inputs
        shutil.rmtree(tune_input_dir, ignore_errors=True)
        # Remove temporary tuning logs
        if uc.rm_tmp_log:
            # tmp_log_dir = pathlib.Path(rc.save_sub_dir) / "tmp_log"
            tmp_log_dir = uc.tmp_dir
            shutil.rmtree(tmp_log_dir, ignore_errors=True)

    def execute_apply_job(self):
        # Set aliases
        jc = self.job_config.job
        rc = self.job_config.run
        ic = self.job_config.input
        tc = self.job_config.train
        ac = self.job_config.apply

        # Setup save parameters if reports need to be saved
        rc.save_dir = f"{rc.save_sub_dir}/apply/{jc.job_name}"
        pathlib.Path(rc.save_dir).mkdir(parents=True, exist_ok=True)
        # Save the final job_config
        yaml_path = pathlib.Path(rc.save_dir) / "apply_config.yaml"
        with open(yaml_path, "w") as yaml_file:
            yaml.dump(self.job_config.get_config_dict(), yaml_file, indent=4)

        # Load inputs
        df_raw = self.model_wrapper.get_feedbox().get_raw_df()
        if logger.level == logging.DEBUG:
            logger.warn(
                f"Randomly sampling 10000 events as input for debugging purpose."
            )
            time.sleep(3)
            df_raw = df_raw.sample(n=10000)
            df_raw.reset_index(drop=True, inplace=True)
        df = self.model_wrapper.get_feedbox().get_processed_df(raw_df=df_raw)

        # Studies not depending on models
        ## metrics curves
        if ac.book_history:
            train_history.plot_history(
                self.model_wrapper, self.job_config, save_dir=rc.save_dir
            )
        ## input kinematic plots
        if ac.book_kine:
            logger.info("Plotting input (raw) distributions.")
            kinematics.plot_input(
                df_raw,
                self.job_config,
                save_dir=f"{rc.save_dir}/kinematics/raw",
                is_raw=True,
            )
            logger.info("Plotting input (processed) distributions.")
            kinematics.plot_input(
                df,
                self.job_config,
                save_dir=f"{rc.save_dir}/kinematics/processed",
                is_raw=False,
            )
        ## correlation matrix
        if ac.book_cor_matrix:
            logger.info("Making correlation matrix")
            kinematics.plot_correlation_matrix(
                df_raw, self.job_config, save_dir=f"{rc.save_dir}/kinematics"
            )

        # Studies depending on models
        if ac.jump_model_studies:
            logger.info("Ignoring model dependent studies as specified")
            return

        ## generate fit arrays
        if ac.book_fit_inputs:
            save_region = ac.fit_df.region
            if save_region is None:
                save_region = ic.region
            logger.info("Dumping numpy arrays for fitting.")
            evaluate_utils.dump_fit_df(
                self.model_wrapper,
                df_raw,
                df,
                self.job_config,
                save_dir=ac.fit_df.save_dir,
            )

        ## loop over models at different epochs
        epoch_checklist = [None]
        if ac.check_model_epoch:
            for epoch_id in range(tc.epochs):
                epoch = epoch_id + 1
                if epoch % ac.epoch_check_interval == 1:
                    epoch_checklist.append(epoch)
        for epoch in epoch_checklist:
            logger.info(">" * 80)
            if epoch is None:
                logger.info(f"Checking model(s) at final epoch")
                epoch_str = "final"
            else:
                logger.info(f"Checking model(s) at epoch {epoch}")
                epoch_str = str(epoch)
            self.model_wrapper.load_model(epoch=epoch)

            # create epoch sub-directory
            epoch_subdir = evaluate_utils.create_epoch_subdir(
                rc.save_dir, epoch, len(str(tc.epochs))
            )
            if epoch_subdir is None:
                logger.error(
                    f"Can't create epoch subdir, skip evaluation at epoch {epoch_str}!"
                )
                return

            # update y values predictions
            y_pred, _, _ = evaluate_utils.k_folds_predict(
                self.model_wrapper.get_model(), df[ic.selected_features].values
            )
            df.loc[:, "y_pred"] = y_pred

            # metrics (PR, ROC, confusion matrix)
            if ac.book_confusion_matrix or ac.book_roc or ac.book_pr:
                metrics.make_metrics_plot(
                    df_raw, df, self.job_config, epoch_subdir
                )
            # overtrain check
            if ac.book_train_test_compare:
                mva_scores.plot_train_test_compare(
                    df, self.job_config, epoch_subdir
                )
            # data/mc scores comparison
            if ac.book_mva_scores_data_mc:
                mva_scores.plot_mva_scores(
                    df_raw, df, self.job_config, epoch_subdir,
                )
            # Make significance scan plot
            if ac.book_significance_scan:
                significance.plot_significance_scan(
                    df, self.job_config, epoch_subdir
                )
            # kinematics with DNN cuts
            if ac.book_cut_kine_study:
                for dnn_cut in ac.cfg_cut_kine_study.dnn_cut_list:
                    dnn_kine_path = (
                        epoch_subdir / f"kine_cut_dnn_p{dnn_cut * 100}"
                    )
                    dnn_kine_path.mkdir(parents=True, exist_ok=True)
                    kinematics.plot_input_dnn(
                        df_raw,
                        df,
                        self.job_config,
                        dnn_cut=dnn_cut,
                        save_dir=dnn_kine_path,
                    )
            # feature permuted importance
            if ac.book_importance_study:
                logger.info("Checking input feature importance")
                importance.plot_feature_importance(
                    self.model_wrapper, df, self.job_config, epoch_subdir
                )

            logger.info("<" * 80)

        return

    def fix_random_seed(self):
        """Fixes random seed

        Ref:
            https://keras.io/getting_started/faq/#how-can-i-obtain-reproducible-results-using-keras-during-development

        Note:
            The seed setting funcions called in this function shouldn't be set 
            again in later code, otherwise extra randomness will be introduced 
            (even if set same seed). The reason is unknown yet.

        """
        seed = self.job_config.job.rdm_seed
        # The below is necessary for starting Numpy generated random numbers
        # in a well-defined initial state.
        np.random.seed(seed)
        # The below is necessary for starting core Python generated random numbers
        # in a well-defined state.
        python_random.seed(seed)
        # The below set_seed() will make random number generation
        # in the TensorFlow backend have a well-defined initial state.
        # For further details, see:
        # https://www.tensorflow.org/api_docs/python/tf/random/set_seed
        tf.random.set_seed(seed)

    def get_config(self, yaml_path: ht.pathlike):
        """Retrieves configurations from yaml file."""
        cfg_path = job_utils.get_valid_cfg_path(yaml_path)
        if not pathlib.Path(cfg_path).is_file():
            logger.error("No vallid configuration file path provided.")
            raise FileNotFoundError
        yaml_dict = config_utils.load_yaml_dict(cfg_path)
        job_config_temp = config_utils.Hepy_Config(yaml_dict)
        # Check whether need to import other (default) ini file first
        if hasattr(job_config_temp, "config"):
            import_ini_path_list = job_config_temp.config.include
            if import_ini_path_list:
                for cfg_path in import_ini_path_list:
                    self.get_config(cfg_path)
                    logger.info(f"Included config: {cfg_path}")
        if self.job_config:
            self.job_config.update(yaml_dict)
        else:
            self.job_config = job_config_temp

        jc = self.job_config.job
        ic = self.job_config.input
        rc = self.job_config.run
        if jc.date_str is None:
            rc.datestr = datetime.date.today().strftime("%Y-%m-%d")
        else:
            rc.datestr = jc.date_str
        if ic.selected_features:
            rc.input_dim = len(ic.selected_features)
        rc.config_collected = True

    def set_model(self):
        logger.info("Setting up model")
        tc = self.job_config.train
        jc = self.job_config.job
        model_class = train_utils.get_model_class(tc.model_class)
        self.model_wrapper = model_class(self.job_config)
        # Load model for "apply" job
        if jc.job_type == "apply":
            self.model_wrapper.load_model()

    def set_model_input(self):
        logger.info("Processing inputs")
        self.model_wrapper.set_inputs(self.job_config)

    def set_save_dir(self, resume: bool = False):
        """Sets the directory to save the outputs"""
        jc = self.job_config.job
        rc = self.job_config.run

        # Determine sub-directory for result saving
        create_new = True
        job_dir_name = jc.job_name
        if jc.job_type == "tune":
            if resume:
                create_new = False
        elif jc.job_type == "train":
            if jc.tune_job_name != "TUNE_JOB_NAME_DEF":
                create_new = False
                job_dir_name = jc.tune_job_name
        elif jc.job_type == "apply":
            create_new = False
            if jc.tune_job_name != "TUNE_JOB_NAME_DEF":
                job_dir_name = jc.tune_job_name
            elif jc.train_job_name != "TRAIN_JOB_NAME_DEF":
                job_dir_name = jc.train_job_name
            else:
                logger.critical(
                    f"Job type is apply but neither tune_job_name nor train_job_name specified! Please update the config."
                )
                exit()
        else:
            logger.critical(f"Unknown job_type {jc.job_type}!")
            exit()

        if create_new:
            dir_pattern = f"{jc.save_dir}/{rc.datestr}_{job_dir_name}_v{{}}"
            output_match = common_utils.get_newest_file_version(dir_pattern)
            rc.save_sub_dir = output_match["path"]
        else:
            if jc.fix_date_str:
                dir_pattern = (
                    f"{jc.save_dir}/{rc.datestr}_{job_dir_name}_v{{}}"
                )
            else:
                dir_pattern = f"{jc.save_dir}/*_{job_dir_name}_v{{}}"
            output_match = common_utils.get_newest_file_version(
                dir_pattern, use_existing=True
            )
            if output_match:
                rc.save_sub_dir = output_match["path"]
            else:
                logger.error(
                    f"Can't find existing work folder matched pattern {dir_pattern}, please check the settings."
                )
                exit()
        logger.info(f"Setup work directory: {rc.save_sub_dir}")
        pathlib.Path(rc.save_sub_dir).mkdir(parents=True, exist_ok=True)

        # Set input cache for tune job
        if jc.job_type == "tune":
            tune_input_cache = pathlib.Path(rc.save_sub_dir) / "tmp"
            tune_input_cache.mkdir(parents=True, exist_ok=True)
            rc.tune_input_cache = str(tune_input_cache)
